/*
===============================================================================
 Name        : MP3_16.c
 Author      : $(author)
 Version     :
 Copyright   : $(copyright)
 Description : main definition
===============================================================================
*/

#ifdef __USE_CMSIS
#include "LPC17xx.h"
#endif

#include <cr_section_macros.h>
#include "stdlib.h"
#include "string.h"
#include "i2s.h"
#include "timer.h"
#include "i2c.h"
#include "type.h"
// libmad
#include "libmad/config.h"
#include "libmad/decoder.h"
// FatFS
#include "usbhost_inc.h"
#include "ff.h"
#include "ffconf.h"
#include "uart.h"
// 自己／デバッグ
#include "mp3_16.h"

#define	USE_I2S_DMA			// I2Sへのデータ転送にDMAを使う

#define SAMPLE_FREQ 			44100	// サンプリング周波数の初期値

#define INPUT_BUF_SIZE			(512*4)		// madlibのinput関数内バッファのサイズ 	2100

#define TIMER0_INTERVAL			((SystemCoreClock/(4*1000*10)))			// 0.1ms (PCLK=CCLK/4)
#define TIMER1_INTERVAL			((SystemCoreClock/(4*5)))				// 200ms=1/5s (PCLK=CCLK/4)
#define TIMER1_1MIN				(5*60)		// 1分
#define TIMER2_INTERVAL			(SystemCoreClock/4)						// 1s (PCLK=CCLK/4)
#define	TIMER2_PRESCALE			20
#define	AUTOPOFF_PAUSE			2			// 自動電源オフ　タイムアウト時間（分）　選択入力待ちのとき
#define	AUTOPOFF_PLAYING		180			// 同　再生中のとき

// ボタン操作
#define SW_PRESSED				0x01
#define SWSELECT_PRESSED      	0x01
#define SWRETURN_PRESSED      	0x02
#define SWDOWN_PRESSED          0x03
#define SWUP_PRESSED            0x04
#define SWOPTION_PRESSED     	0x05
#define SWPLAY_PRESSED          0x06
#define SWRANDOM_PRESSED        0x07
#define SWPAGEDOWN_PRESSED      0x08
#define SWPAGEUP_PRESSED        0x09
#define SWPOFF_PRESSED			0x0A

#define POWER_ON_BIT			30		// P1.30
#define	SPEAKER_OR_EARPHONE		26		// P0.26
#define SPEAKER_MUTE_BIT		25		// P0.25

#define DEBUG_MSG   DebugWriteln

// OLED
#define PORT_USED			1
#define OLED_ADDRESS        0x3C
#define OLED_CB_COMMAND     0x00
#define OLED_CB_COMMAND_CO  (0x00 | 0x80)
#define OLED_CB_DATA        0x40
#define OLED_CB_DATA_CO     (0x40 | 0x80)
#define OLED_WIDTH			128

#define LCD_BYTE_HEIGHT			8		// 8バイト高=64ドット
#define	LCD_IMAGE_BUFFER_SIZE	(LCD_BYTE_HEIGHT * OLED_WIDTH)
#define LCD_STRING_LINES		4		// LCDに表示する文字列行数(固定値)
#define MAX_SELDIR_ITEM			4		// DIR選択画面で一度に表示されるDIR数の最大数
#define MAX_SELMP3_ITEM         4       // 曲選択画面で一度に表示される数

// DIR構成ファイル　レコード長・項目別オフセット＆バイトサイズ
#define	DIRSTRUCTUREFILE				"/DIRLST16.DAT"		// DIR構成ファイル(Version=16)
#define	RECLEN_DIRLIST					512UL				// イメージ部分を除くレコード長
#define	RECLEN_DIRLIST_FULL				(512UL + 512UL)		// イメージ部分を含むレコード長
#define DL_DISPLAYNAME_OFFSET			0					// ディレクトリ表示名
#define DL_DISPLAYNAME_LEN				32
#define DL_FULLDIR_OFFSET				(DL_DISPLAYNAME_OFFSET + DL_DISPLAYNAME_LEN)	// ディレクトリフルパス
#define DL_FULLDIR_LEN					258
#define DL_FULLDIRBYTES_OFFSET			(DL_FULLDIR_OFFSET + DL_FULLDIR_LEN)			// ディレクトリフルパスバイト数
#define DL_FULLDIRBYTES_LEN				2
#define DL_DIRNAME_OFFSET				(DL_FULLDIRBYTES_OFFSET + DL_FULLDIRBYTES_LEN)	// ディレクトリ名
#define DL_DIRNAME_LEN					130
#define DL_RECNO_SUBDIR_START_OFFSET	(DL_DIRNAME_OFFSET + DL_DIRNAME_LEN)			// サブディレクトリの先頭レコード番号
#define DL_RECNO_SUBDIR_START_LEN		2
#define DL_NUM_SUBDIR_OFFSET			(DL_RECNO_SUBDIR_START_OFFSET + DL_RECNO_SUBDIR_START_LEN)		// サブディレクトリ数
#define DL_NUM_SUBDIR_LEN				1
#define DL_RECNO_PARENTDIR_OFFSET		(DL_NUM_SUBDIR_OFFSET + DL_NUM_SUBDIR_LEN)		// 親ディレクトリのレコード番号
#define DL_RECNO_PARENTDIR_LEN			2
#define DL_RECNO_GRANDPARENTDIR_OFFSET	(DL_RECNO_PARENTDIR_OFFSET + DL_RECNO_PARENTDIR_LEN)	// 親々ディレクトリのレコード番号
#define DL_RECNO_GRANDPARENTDIR_LEN		2
#define DL_NUM_SNDFILE_OFFSET			(DL_RECNO_GRANDPARENTDIR_OFFSET + DL_RECNO_GRANDPARENTDIR_LEN)		// MP3ファイル数
#define DL_NUM_SNDFILE_LEN				1
#define DL_RECNO_SNDFILE_OFFSET			(DL_NUM_SNDFILE_OFFSET + DL_NUM_SNDFILE_LEN)	// MP3ファイルリスト先頭レコード番号
#define DL_RECNO_SNDFILE_LEN			2
#define DL_FILLER_LEN					80
#define DL_IMAGEDATA_OFFSET				RECLEN_DIRLIST		// イメージデータ
#define DL_IMAGEDATA_LEN				512
#define DL_IMAGEDATAHALF_LEN			256

// 音源ファイルリスト
#define	SNDFILELISTFILE					"/SNDLST16.DAT"		// MP3ファイルリスト(Version=16)
#define	RECLEN_MP3LIST					352UL					// イメージ部分を除くレコード長
#define	RECLEN_MP3LIST_FULL				1120UL					// レコード長
#define ML_DISPLAYNAME_OFFSET			0
#define ML_DISPLAYNAME_LEN				48
#define ML_FILENAME_OFFSET				(ML_DISPLAYNAME_OFFSET + ML_DISPLAYNAME_LEN)
#define ML_FILENAME_LEN					258
#define ML_FILLER_LEN					46
#define ML_IMAGEDATA_OFFSET				RECLEN_MP3LIST
#define ML_IMAGEDATA_LEN				768UL
#define ML_IMAGEDATAHALF_LEN			384UL

// ランダム再生順ファイル
#define	RANDOMORDERFILE					"/RANDOM16.DAT"
#define	RECLEN_RANDOMLIST				4
#define	RD_UNIT_OFFSET					0
#define	RD_UNIT_LEN						1
#define	RD_DIRRECNO_OFFSET				(RD_UNIT_OFFSET + RD_UNIT_LEN)
#define	RD_DIRRECNO_LEN					2
#define	RD_TRACKNO_OFFSET				(RD_DIRRECNO_OFFSET + RD_DIRRECNO_LEN)
#define	RD_TRACKNO_LEN					1

// 履歴ファイル HISTFILE parameters
#define HISTFILE                		"/HIST16.DAT"
#define RECLEN_HISTFILE         		10UL
#define HS_LASTDIR_OFFSET       		0
#define HS_LASTDIR_LEN	       			2
#define HS_LASTTRACK_OFFSET				(HS_LASTDIR_OFFSET + HS_LASTDIR_LEN)
#define HS_LASTTRACK_LEN				1
#define HS_LASTBYTE_OFFSET				(HS_LASTTRACK_OFFSET + HS_LASTTRACK_LEN)
#define HS_LASTBYTE_LEN             	4
#define	HS_RANDOMREC_OFFSET				(HS_LASTBYTE_OFFSET + HS_LASTBYTE_LEN)
#define	HS_RANDOMREC_LEN				2
#define	HS_OUTPUT_OFFSET				(HS_RANDOMREC_OFFSET + HS_RANDOMREC_LEN)
#define	HS_OUTPUT_LEN					1

// ボリュームファイル VOLUMEFILE parameters
#define VOLUMEFILE              		"VOLUME16.DAT"		// ディレクトリごとに作成
#define	VF_VOLIDX_EAR_OFFSET			0												// Earphone時のVolumeIdx
#define	VF_VOLIDX_EAR_LEN				1
#define VF_FACTOR_EAR_OFFSET			(VF_VOLIDX_EAR_OFFSET + VF_VOLIDX_EAR_LEN)		// Earphone時のVolumeFactor
#define VF_FACTOR_EAR_LEN				sizeof(mad_fixed_t)
#define	VF_VOLIDX_SPK_OFFSET			(VF_FACTOR_EAR_OFFSET + VF_FACTOR_EAR_LEN)		// Speaker時のVolumeIdx
#define	VF_VOLIDX_SPK_LEN				1
#define VF_FACTOR_SPK_OFFSET			(VF_VOLIDX_SPK_OFFSET + VF_VOLIDX_SPK_LEN)		// Speaker時のVolumeFactor
#define VF_FACTOR_SPK_LEN          		sizeof(mad_fixed_t)
#define RECLEN_VOLUME           		(VF_FACTOR_SPK_OFFSET + VF_FACTOR_SPK_LEN)

// メッセージファイル
#define MSGSTRFILE						"/MSGSTR16.DAT"
#define MS_IMAGEDATA_LEN				256UL
#define MS_IMAGEDATAHALF_LEN			128UL
#define	MSGNO_MP3_16					0
#define	MSGNO_SPEAKER					1
#define	MSGNO_EARPHONE					2
#define	MSGNO_DIRLST_ERR				3
#define	MSGNO_SNDLST_ERR				4
#define	MSGNO_RANDOM_ERR				5
#define	MSGNO_MP3_ERR					6
#define	MSGNO_NO_HISTFILE				7
#define	MSGNO_SEL_RET					8

// ファイル読み込み用バッファサイズ
#if RECLEN_DIRLIST_FULL > RECLEN_MP3LIST_FULL
	#define	REC_BUF_SIZE	RECLEN_DIRLIST_FULL
#else
	#define	REC_BUF_SIZE	RECLEN_MP3LIST_FULL
#endif

/*----------------------------------------------------------------------------
  FAT
 *----------------------------------------------------------------------------*/
FATFS Fatfs[_VOLUMES];		/* File system object for each logical drive */
uint8_t Drives;				/* Drives mask */
FIL FilMP3;

/*----------------------------------------------------------------------------
  MP3 Decoder
 *----------------------------------------------------------------------------*/
#define  AUDIOBUF_SIZE		(1152*2)

/* Allocate 2 or 3 audio buffer to store PCM after decoding */
#define  MAX_AUDIOBUF_NUM	2
#if (MAX_AUDIOBUF_NUM < 2) || (MAX_AUDIOBUF_NUM > 3)
    #error Invalid MAX_AUDIOBUF_NUM setting.
#endif

#define	AUDIOBUF_EMPTY		1		// audiobufにデータなし
#define	AUDIOBUF_READY		0		// audiobufにデータあり（再生中または再生待ち）

enum {AUDIOBUF0=0, AUDIOBUF1=1};
typedef struct
{
	uint32_t BaseAddr;  /* base address for the audio buf */
	int32_t Size;		/* size(Byte) of the audio buf */
	uint8_t Empty;		/* buf status, 1-> empty */
} AUDIOBUF;

AUDIOBUF audiobuf[MAX_AUDIOBUF_NUM] = {
	{(uint32_t)NULL, -1, AUDIOBUF_EMPTY},
	{(uint32_t)NULL, -1, AUDIOBUF_EMPTY},
#if MAX_AUDIOBUF_NUM==3
	{(uint32_t)NULL, -1, AUDIOBUF_EMPTY},
#endif
};

typedef unsigned short		dacout_s;
dacout_s dacbuf[AUDIOBUF_SIZE*2 /*1152*4*/];
dacout_s *dac_s, *dac_e;
volatile uint8_t PlayingBufIdx = AUDIOBUF0;
volatile uint8_t DecodeBufIdx = AUDIOBUF0;		// dacflgに代わる
void	*pDecoderSync;		// decoder->sync 退避 (decoder.c)
void	*pStreamMainData;	// stream->main_data退避 (layer3.c)
void	*pFrameOverlap;		// frame->overlap退避 (layer3.c)
unsigned int samplerate;	// サンプルレートの退避
int mono;					// モノ／ステレオの退避
enum mad_flow (*output_func_backup)(void *, struct mad_header const *, struct mad_pcm *);	// output_funcの一時的退避

/*----------------------------------------------------------------------------
  I2S and DMA
 *----------------------------------------------------------------------------*/
#if defined(USE_I2S_DMA)

/* Wait for DMA is completed. */
#define WAIFORDMADONE   while(I2S_DMA_Done == 0)
/* Wait for I2S TX buffer is empty. */
#define WAIFORI2SDONE	while (((LPC_I2S->I2SSTATE>>16) & 0xF) != 0)

/* GPDMA for I2S */
#define I2S_DMA_CHAN	7
#if I2S_DMA_CHAN<0 || I2S_DMA_CHAN>8
	#error Invalid DMA channel number!
#endif
#define DMA_I2S_TX_FIFO	0x400A8008

#define RENAME(x,y)  RENAME2(x,y)
#define RENAME2(x,y)	x##y

LPC_GPDMACH_TypeDef *LPC_GPDMACH = RENAME(LPC_GPDMACH, I2S_DMA_CHAN);

volatile uint8_t I2S_DMA_Done = 0;
volatile uint8_t DMA_Buf_Empty = 0;

#endif		// defined(USE_I2S_DMA)

/*----------------------------------------------------------------------------
  その他
 *----------------------------------------------------------------------------*/
// イベント/制御
static volatile int	PinEvent;
static volatile uint16_t	CHATTERING_TMR;
static volatile uint16_t	AutoPoffTimer;
static uint8_t		ChannelMode;	// 0=STEREO, 1=MONO
static uint8_t		NowPlaying;
static uint8_t		RandomPlaying;	// 1=ランダム再生中
static uint8_t    	LcdImageBuff[LCD_IMAGE_BUFFER_SIZE]; 	// LCDフル画面イメージデータ
static uint8_t      RecHistory[RECLEN_HISTFILE];			// 履歴ファイルレコード

typedef enum {PD_SEQUENTIAL=0, PD_ONLY_ONE } playdirectory_t;		// ディレクトリ下の音源を順次再生するか、特定のファイルだけを再生するか

// 音量
static mad_fixed_t		VolumeFactor;
static int 				VolumeIdx;
//static uint16_t		VolumeTab[] = {8,10,13,16,20,25,32,40,50,63,
//									79,100,126,158,199,251,316,398,501,630,
//									793,999,1258,1583,1994,2510,3160,3978,5009,6306};	// MAX_VOLUME=10000
//#define	DEF_VOLUMETAB_IDX_EAR	(MAX_VOLUMETAB_IDX - 6)			// 1583 ear-phone
//#define	DEF_VOLUMETAB_IDX_SPK	(MAX_VOLUMETAB_IDX - 15)		// 199  speaker
static uint16_t		VolumeTab[] = {0, 8, 10, 11, 12, 14, 15, 17, 19, 22, 25,	// 0-10
 28, 31, 35, 39, 44, 50, 56, 63, 70, 79,			// 11-20
 89, 100, 112, 125, 141, 158, 177, 199, 223, 251,	// 21-30
 281, 316, 354, 398, 446, 501, 562, 630, 707, 794,	// 31-40
 891, 1000, 1122, 1258, 1412, 1584, 1778, 1995, 2238, 2511,		// 41-50
 2818, 3162, 3548, 3981, 4466, 5011, 5623, 6309, 7079, 7943,	// 51-60
 8912, 10000 };		// 61-62
#define	MIN_VOLUMETAB_IDX		0
#define MAX_VOLUMETAB_IDX		((sizeof(VolumeTab)/sizeof(VolumeTab[0])) - 1)
#define	DEF_VOLUMETAB_IDX_EAR	(MAX_VOLUMETAB_IDX - 17)		// ear-phone
#define	DEF_VOLUMETAB_IDX_SPK	(MAX_VOLUMETAB_IDX - 34)		// speaker
#define MIN_VOLUME              0		// Volume最小値
#define MAX_VOLUME              10000	// Volume最大値
#define	VOLUMEIDX_STEP			3
#define	OUTPUT_SPEAKER			1		// 出力先
#define OUTPUT_EARPHONE			0

// I2C/OLED
extern volatile uint8_t I2CMasterBuffer[I2C_PORT_NUM][I2C_BUFSIZE];
extern volatile uint32_t I2CReadLength[I2C_PORT_NUM];
extern volatile uint32_t I2CWriteLength[I2C_PORT_NUM];

// 関数宣言
static int decode();
void SystemInit (void);
void SystemCoreClockUpdate (void);
uint8_t Mem_Init (void);
int USBHost_Init(void);
void InitApp();
int PlayOneMp3(const TCHAR *mp3file, uint16_t dirno, uint8_t trackno, uint32_t start_byte);
void PowerOff();
void SpeakerAmpEnable();
void SpeakerAmpDisable();
void SetOutput();
void ActivateSpeaker(int onoff);
int SelectDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3);
int SelectSoundfile(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno);
int PlayOneDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno, uint32_t start_byte, playdirectory_t mode);
int PlayRandom(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3);
int SelectOption(FIL *Fil);
int PauseAndSelectItem();
int ReadHistory(FIL *Fil);
int SaveHistory(FIL *Fil);
int SetLocalVolume(const TCHAR *volumefile, int output, FIL *Fil);
int SaveVolumeFile(const TCHAR *volumefile, int output, FIL *Fil);
#if defined(USE_I2S_DMA)
void DMA_Init(void);
void DMA_SendData(uint32_t *buf, uint32_t size);
#else
void SysTick_Off();
void SysTick_On();
#endif
void InitOled();
void DispLine(int page, uint8_t *imagedata);
void SetMenuItemSelected(uint8_t line);
void DispLcdFullImage();
void ClearLcd();
void DispMsg(int line, int msgno, int inv, FIL *Fil);
void OledTest();
void LFN_play_test();
void LFN_play_test2();
WCHAR *MakeUnicodeString(char *s);
void WaitMs(int ms);
void Led2_On();
void Led2_Off();
void Led3_On();
void Led3_Off();
void Led4_On();
void Led4_Off();
//void test_read_write();


int main(void) {
    uint8_t     RecBuf[REC_BUF_SIZE];
    FIL 		FilDL;
    FIL 		FilML;
    FIL 		Fil3;
    int			ret;
    FRESULT     res;

    SystemInit();
    SystemCoreClockUpdate();
    InitApp();
    CHATTERING_TMR = 0;
    AutoPoffTimer = 0;
    NowPlaying = 0;
    RandomPlaying = 0;
    DispMsg(0, MSGNO_MP3_16, 0, &Fil3);
    ReadHistory(&Fil3);
    if (RecHistory[HS_OUTPUT_OFFSET] == OUTPUT_SPEAKER){
        DispMsg(1, MSGNO_SPEAKER, 0, &Fil3);
    } else {
        DispMsg(1, MSGNO_EARPHONE, 0, &Fil3);
    }
    WaitMs(2000);
    SetOutput();

    //InitOled();
	//OledTest();

//    test_read_write();

//	LFN_play_test();
//	LFN_play_test2();

    // DIR構成ファイル
    res = f_open(&FilDL, MakeUnicodeString(DIRSTRUCTUREFILE), FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, &Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_open error\n\r");
        goto POFF;
    }
    // MP3ファイルリスト
    res = f_open(&FilML, MakeUnicodeString(SNDFILELISTFILE), FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_SNDLST_ERR, 0, &Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_open error\n\r");
        goto POFF;
    }

    do {
		ret = SelectDirectory(RecBuf, &FilDL, &FilML, &Fil3);
		if (ret == START_RANDOM){
			ret = PlayRandom(RecBuf, &FilDL, &FilML, &Fil3);
			if (ret == POWEROFF){
				break;
			}
		} else break;
    } while (1);

    f_close(&FilDL);
    f_close(&FilML);
//    Led2_Off();
//    Led3_Off();
//    Led4_Off();

    POFF:
	DEBUG_MSG("MP3_16 end.\n\r");
    PowerOff();
    // Enter an infinite loop, just incrementing a counter
    while(1) {
//        i++ ;
    }
    return 0 ;
}

/*
 * 各種初期化(IO/Timer/I2S/SPI/I2C/UART/GCLD/メモリ領域)
 */
void InitApp(){
    LPC_GPIO1->FIODIR |= (0x01 << POWER_ON_BIT);		// P1.30 (output)
    LPC_GPIO1->FIOSET |= (0x01 << POWER_ON_BIT);		// P1.30 (set high)
    LPC_GPIO1->FIODIR |= ((0x01 << 18) | (0x01 << 20) | (0x01 << 21) | (0x01 << 23));		// on board led
    LPC_GPIO0->FIODIR |= (0x01 << SPEAKER_MUTE_BIT);	// P0.26 (output)
    SpeakerAmpDisable();
    LPC_GPIO0->FIODIR |= (0x01 << SPEAKER_OR_EARPHONE);	// P0.25 (output)

    // power swがリリースされるのを待つ
    while ((LPC_GPIO1->FIOPIN & (1 << 31)) == 0){	// P1.31

    }
	init_timer( 0, TIMER0_INTERVAL );	// USBメモリ検出のタイムアウトに使う
	enable_timer( 0 );
    UARTInit(1, 9600);
	DEBUG_MSG("MP3_16 start.\n\r");
	I2C1Init();
    InitOled();

    Drives = Mem_Init();
    if (Drives == 0)
    {
        DEBUG_MSG ("Memory: Init failed.\n\r");
        return;
    }

	i2s_init();		// i2s.c
	init_timer( 1, TIMER1_INTERVAL );
	enable_timer( 1 );

	pDecoderSync = 0;
	pStreamMainData = 0;
	pFrameOverlap = 0;

	audiobuf[0].BaseAddr = dacbuf;
	audiobuf[1].BaseAddr = &dacbuf[AUDIOBUF_SIZE];
    DecodeBufIdx = PlayingBufIdx = AUDIOBUF0;
    audiobuf[AUDIOBUF0].Empty = audiobuf[AUDIOBUF1].Empty = AUDIOBUF_EMPTY;
    dac_s = dac_e = audiobuf[PlayingBufIdx].BaseAddr;

	//VolumeFactor = MAD_F_ONE;		//mad_f_tofixed((long)options.volume/100.0);
    VolumeIdx = DEF_VOLUMETAB_IDX_EAR;	// 暫定値。再生時にディレクトリごとに値を設定する。
	VolumeFactor = mad_f_tofixed((long)VolumeTab[VolumeIdx]/(float)MAX_VOLUME);

	mono = I2S_STEREO;
	samplerate = SAMPLE_FREQ;
	i2s_stereomono(mono);
	i2s_begin(samplerate, 16);

#if defined(USE_I2S_DMA)
    DMA_Init(); /* Init DMA after init I2S */
#endif

	PinEvent = 0;
}

/**
  * @brief  Initializes memory.
  *
  * @param  None.
  * @retval Mask of mounted drives
  *
  */
uint8_t Mem_Init (void)
{
    uint8_t retv = 0x0;
//    static TCHAR drv[] = {0x0030, 0x003A, 0x0000};

	/* Wait until there is USB flash disk inserted. */
	if (USBHost_Init() == OK)
	{
		DEBUG_MSG("USBHost: Init OK.\n\r");
//		if (f_mount(/*pFatfs*/ &Fatfs[0], drv, 0) == FR_OK){
		if (f_mount(/*pFatfs*/ &Fatfs[0], MakeUnicodeString("0:"), 0) == FR_OK){
//		if (f_mount(/*pFatfs*/ &Fatfs[0], "0:", 0) == FR_OK){
			DEBUG_MSG("USBFS: Mount OK.\n\r");
			retv |= 1;
		} else {
			DEBUG_MSG("USBFS: Mount failed.\n\r");
	    }
	}
	else
	{
		DEBUG_MSG("USBHost: Init failed.\n\r");
		return retv;
	}

#ifdef SUPPORT_SD
	if (f_mount(1, &Fatfs[1]) == FR_OK) {
		DEBUG_MSG("SDFS: Mount OK.\n\r");
		LCD_PrintString(0*8, 6*15, "SDFS: Mount OK", COLOR_LIME, COLOR_BLACK);
		retv |= 2;
	} else {
		DEBUG_MSG("SDFS: Mount failed.\n\r");
		LCD_PrintString(0*8, 6*15, "SDFS: mnt fail", COLOR_RED, COLOR_BLACK);
    }
#endif

    return (retv);
}

/*----------------------------------------------------------------------------
  USB functions
 *----------------------------------------------------------------------------*/
/**
  * @brief  Initializes the USBHost .
  *
  * @param  None
  * @retval 0:OK, others: failure code
  *
  */
int USBHost_Init(void)
{
    int32_t  rc;
	uint32_t  numBlks, blkSize;
	uint8_t  inquiryResult[INQUIRY_LENGTH];

    Host_Init();               /* Initialize the lpc17xx host controller */
    rc = Host_EnumDev();       /* Enumerate the device connected  */
	if (rc != OK)
		return 1;
	rc = MS_Init( &blkSize, &numBlks, inquiryResult );
	if ((rc != OK) || (blkSize != 512))
		return 2;

	return OK;
}


/*
 * MP3再生
 * 	mp3file:MP3ファイルフルパス、dirno:ルートディレクトリ=0から始まる行番号
 */
int PlayOneMp3(const TCHAR *mp3file, uint16_t dirno, uint8_t trackno, uint32_t start_byte){
    FRESULT     res;
    int         ret;

    ret = FR_OK;
	//DEBUG_MSG("PlayOneMp3 start.\n\r");
	//DEBUG_MSG((char *)(mp3file));
	res = f_open(&FilMP3 /*pfilMP3*/, mp3file, FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DEBUG_MSG("f_open ERR.\n\r");
		return res;
	}
	//DEBUG_MSG("f_open OK.\n\r");

	// 途中再開
	if (start_byte > 0){
		if (f_lseek(&FilMP3, start_byte) != FR_OK){      // go to restart point
			f_lseek(&FilMP3, 0);
		} else if (f_eof(&FilMP3)){
			f_close(&FilMP3);
			return ret;
		}
	}
//	ActivateSpeaker(1);
//	//f_lseek(pfilMP3, 4*32*100000L);		// 途中から再生テスト

    memcpy(RecHistory + HS_LASTDIR_OFFSET, &dirno, HS_LASTDIR_LEN);
    *(uint16_t *)(RecHistory + HS_LASTTRACK_OFFSET) = trackno;
    *(uint32_t *)(RecHistory + HS_LASTBYTE_OFFSET) = f_tell(&FilMP3);

//	i2s_init();		// i2s.c
	audiobuf[0].BaseAddr = dacbuf;
	audiobuf[1].BaseAddr = &dacbuf[AUDIOBUF_SIZE];
    DecodeBufIdx = PlayingBufIdx = AUDIOBUF0;
    audiobuf[AUDIOBUF0].Empty = audiobuf[AUDIOBUF1].Empty = AUDIOBUF_EMPTY;		//1;
    audiobuf[AUDIOBUF0].Size = audiobuf[AUDIOBUF1].Size = 0;
    dac_s = dac_e = audiobuf[PlayingBufIdx].BaseAddr;
#if defined(USE_I2S_DMA)
    I2S_DMA_Done = 0;
    DMA_Buf_Empty = 1;		// output関数でDMA転送する
#endif
	NowPlaying = 1;
	ret = decode();			// 戻り値はrun_sync(decoder.c)参照
#if defined(USE_I2S_DMA)
	/* Playback complete */
    WAIFORDMADONE;
	WAIFORI2SDONE;
#endif
	NowPlaying = 0;

    *(uint32_t *)(RecHistory + HS_LASTBYTE_OFFSET) = f_tell(&FilMP3);
//	if (f_eof(&FilMP3) == 0){
//	    *(uint32_t *)(RecHistory + HS_LASTBYTE_OFFSET) = f_tell(&FilMP3);
//	} else {	// 曲の最後まで再生したとき
//	    *(uint32_t *)(RecHistory + HS_LASTBYTE_OFFSET) = 0;		// 曲の先頭に
//	}
	f_close(&FilMP3);
//	ActivateSpeaker(0);

//	SaveHistory(&FilMP3);	// ここで履歴保存すると、次の曲の再生のMP3データ読み出しに影響あり！？
	//DebugWritelnV1("Volume=", VolumeTab[VolumeIdx]);

	return ret;
}

/*
 * This is a private message structure. A generic pointer to this structure
 * is passed to each of the callback functions. Put here any data you need
 * to access from within the callbacks.
 */
struct buffer {
  unsigned char const *start;
  unsigned long length;
};


/*
 * This is the input callback. The purpose of this callback is to (re)fill
 * the stream buffer which is to be decoded. In this example, an entire file
 * has been mapped into memory, so we just call mad_stream_buffer() with the
 * address and length of the mapping. When this callback is called a second
 * time, we are finished decoding.
 */
static
enum mad_flow input(void *data,
		    struct mad_stream *stream)
{
	static unsigned char strmbuff[INPUT_BUF_SIZE];
	int ret;
	int rsz;
	unsigned char *bp;
	UINT	bytes;
	//int		i;
	//uint16_t	vstep;
	//DEBUG_MSG("input.\n\r");

	bp = strmbuff;
	rsz = 0;
	if(stream->error == MAD_ERROR_BUFLEN || stream->buffer == 0 /*NULL*/)
	{
	if(stream->next_frame != 0 /*NULL*/)
	  {
		 rsz = stream->bufend - stream->next_frame;
//		 memmove(pStrmBuff, stream->next_frame, rsz);
		 memmove(strmbuff, stream->next_frame, rsz);
		 bp = strmbuff + rsz;
//		 bp = pStrmBuff + rsz;
	  }
	}
	//struct buffer *buffer = data;
	  //if (!buffer->length)

	//DebugWritelnV1("read bytes.\n\r", INPUT_BUF_SIZE - rsz);
	ret = f_read(&FilMP3 /*pfilMP3*/, bp, INPUT_BUF_SIZE - rsz, &bytes);
	if (ret != FR_OK || bytes == 0){
		DebugWritelnV1("input f_read err.\n\r", ret);
		return MAD_FLOW_STOP;
	}
	//DEBUG_MSG("f_read.\n\r");

	mad_stream_buffer(stream, strmbuff, bytes + rsz);
//	mad_stream_buffer(stream, pStrmBuff, bytes + rsz);
	//mad_stream_buffer(stream, buffer->start, buffer->length);

	//buffer->length = 0;

	if (AutoPoffTimer >= AUTOPOFF_PLAYING){
		PinEvent = 0;
		return MAD_FLOW_POWEROFF;
	}
	if (PinEvent > 0){
		switch (PinEvent){
		case SWPLAY_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_PLAY;
			break;
		case SWRANDOM_PRESSED:
			PinEvent = 0;
			if (RandomPlaying == 1) return MAD_FLOW_RANDOM;		// 再生中にRANDOMボタンが使えるのはランダム再生中だけ
			break;
		case SWRETURN_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_RETURN;			// MP3_16
			//return MAD_FLOW_DESELECT;		// MP3_12
			break;
		case SWOPTION_PRESSED:
			break;
		case SWPOFF_PRESSED:
			PinEvent = 0;
			return MAD_FLOW_POWEROFF;
			break;
		}
		PinEvent = 0;
	}
	return MAD_FLOW_CONTINUE;
}

/*
 * The following utility routine performs simple rounding, clipping, and
 * scaling of MAD's high-resolution samples down to 16 bits. It does not
 * perform any dithering or noise shaping, which would be recommended to
 * obtain any exceptional audio quality. It is therefore not recommended to
 * use this routine if high-quality output is desired.
 * mpg321-0.3.2.origのaudio_linear_ditherを使ってみた。
 */
static inline
signed int scale(mad_fixed_t sample)
{
  /* round */
  sample += (1L << (MAD_F_FRACBITS - 16));

  /* clip */
  if (sample >= MAD_F_ONE)
    sample = MAD_F_ONE - 1;
  else if (sample < -MAD_F_ONE)
    sample = -MAD_F_ONE;

  /* quantize */
  return sample >> (MAD_F_FRACBITS + 1 - 16);
}

/*
 * mpg321-0.3.2.origから引用
 */
struct audio_dither {
  mad_fixed_t error[3];
  mad_fixed_t random;
};

/*
 * NAME:        prng()
 * DESCRIPTION: 32-bit pseudo-random number generator
 * mpg321-0.3.2.origから引用
 */
/* 9/23 static*/ inline
unsigned long prng(unsigned long state)
{
  return (state * 0x0019660dL + 0x3c6ef35fL) & 0xffffffffL;
}

/*
 * NAME:        audio_linear_dither()
 * DESCRIPTION: generic linear sample quantize and dither routine
 * mpg321-0.3.2.origから引用
 */
inline
signed long audio_linear_dither(unsigned int bits, mad_fixed_t sample,
                                struct audio_dither *dither)
{
  unsigned int scalebits;
  mad_fixed_t output, mask, random;

  enum {
    MIN = -MAD_F_ONE,
    MAX =  MAD_F_ONE - 1
  };

  /* noise shape */
  sample += dither->error[0] - dither->error[1] + dither->error[2];

  dither->error[2] = dither->error[1];
  dither->error[1] = dither->error[0] / 2;

  /* bias */
  output = sample + (1L << (MAD_F_FRACBITS + 1 - bits - 1));

  scalebits = MAD_F_FRACBITS + 1 - bits;
  mask = (1L << scalebits) - 1;

  /* dither */
  random  = prng(dither->random);
  output += (random & mask) - (dither->random & mask);

  dither->random = random;

  /* clip */
  if (output > MAX) {
    output = MAX;

    if (sample > MAX)
      sample = MAX;
  }
  else if (output < MIN) {
    output = MIN;

    if (sample < MIN)
      sample = MIN;
  }

  /* quantize */
  output &= ~mask;

  /* error feedback */
  dither->error[0] = sample - output;

  /* scale */
  return output >> scalebits;
}

/*
 * This is the output callback function. It is called after each frame of
 * MPEG audio data has been completely decoded. The purpose of this callback
 * is to output (or play) the decoded PCM audio.
 * ステレオ用
 */
static
enum mad_flow output(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  static struct audio_dither dither;
  dacout_s *dac_w;

  /* pcm->samplerate contains the sampling frequency */
  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  // play back 終了を待つ
  while (audiobuf[DecodeBufIdx].Empty == AUDIOBUF_READY){
//	  Led2_On();
//	  Pause();
  }
//  Led2_Off();

  dac_w = audiobuf[DecodeBufIdx].BaseAddr;
  while (nsamples--) {
		*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*left_ch, VolumeFactor), &dither);
	    dac_w++;
    	*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*right_ch, VolumeFactor), &dither);;
	    dac_w++;
	    left_ch++;
	    right_ch++;
	    audiobuf[DecodeBufIdx].Size += 4;
  }
  audiobuf[DecodeBufIdx].Empty = AUDIOBUF_READY;

  //DebugWritelnV1("output", audiobuf[DecodeBufIdx].Size);

#if defined(USE_I2S_DMA)
  if (DMA_Buf_Empty == 1){
	  PlayingBufIdx = DecodeBufIdx;
	  I2S_DMA_Done = 0;
	  DMA_Buf_Empty = 0;
	  DMA_SendData((uint32_t *)audiobuf[PlayingBufIdx].BaseAddr, audiobuf[PlayingBufIdx].Size);
//	  Led3_Off();
  }
#endif

  DecodeBufIdx++;
  if (DecodeBufIdx == MAX_AUDIOBUF_NUM)
	  DecodeBufIdx = 0;

#if 0		// madplayerの原型
//  LPC_GPIO1->FIOSET |= (0x01 << 18);
	enable_timer( 0 );
  while(dac_s < dac_e){
	  Pause();
	  //int k;
	  //for (k=0; k<100; k++);
	  // wait_us(1);		// dacout関数で全データ出力されるのを待つ
  }
	disable_timer( 0 );
//  LPC_GPIO1->FIOCLR |= (0x01 << 18);
  // 音が途切れることは無いのか？
  dac_e = dacbuf;  // potential thread problem ??  no...
  dac_s = dacbuf;

  while (nsamples--) {
	    signed int sample_l,sample_r;
	    sample_l = scale(*left_ch);
	    sample_r = scale(*right_ch);
		*dac_e = (unsigned short)(sample_l);
	    dac_e++;
		*dac_e = (unsigned short)(sample_r);
	    dac_e++;
	    left_ch++;
	    right_ch++;
  }
#endif

  return MAD_FLOW_CONTINUE;
}

/*
 * MONO化output関数
 */
static
enum mad_flow output_monauralize(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
	  unsigned int nsamples;
	  mad_fixed_t const *left_ch, *right_ch;
	  mad_fixed_t	mono_w;
	  static struct audio_dither dither;
	  dacout_s *dac_w;

	  /* pcm->samplerate contains the sampling frequency */
	  //nchannels = pcm->channels;
	  nsamples  = pcm->length;
	  left_ch   = pcm->samples[0];
	  right_ch  = pcm->samples[1];

	  // play back 終了を待つ
	  while (audiobuf[DecodeBufIdx].Empty == AUDIOBUF_READY){
		  Led2_On();
//		  Pause();
	  }
	  Led2_Off();

	  dac_w = audiobuf[DecodeBufIdx].BaseAddr;
	  while (nsamples--) {
		 mono_w = *left_ch / 2 + *right_ch / 2;
			*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(mono_w, VolumeFactor), &dither);
			dac_w++;
	    	*dac_w = *(dac_w - 1);
			dac_w++;
			left_ch++;
			right_ch++;
		    audiobuf[DecodeBufIdx].Size += 4;
	  }
	  audiobuf[DecodeBufIdx].Empty = AUDIOBUF_READY;
#if defined(USE_I2S_DMA)
	  if (DMA_Buf_Empty == 1){
		  PlayingBufIdx = DecodeBufIdx;
		  I2S_DMA_Done = 0;
		  DMA_Buf_Empty = 0;
		  DMA_SendData((uint32_t *)audiobuf[PlayingBufIdx].BaseAddr, audiobuf[PlayingBufIdx].Size);
		  Led3_Off();
	  }
#endif
	  DecodeBufIdx++;
	  if (DecodeBufIdx == MAX_AUDIOBUF_NUM)
		  DecodeBufIdx = 0;

	  return MAD_FLOW_CONTINUE;
}


/*
 * MONO用output関数
 */
static
enum mad_flow output_mono(void *data,
		     struct mad_header const *header,
		     struct mad_pcm *pcm)
{
  unsigned int nsamples;
  mad_fixed_t const *left_ch, *right_ch;
  static struct audio_dither dither;
  dacout_s *dac_w;

  //nchannels = pcm->channels;
  nsamples  = pcm->length;
  left_ch   = pcm->samples[0];
  right_ch  = pcm->samples[1];

  // play back 終了を待つ
  while (audiobuf[DecodeBufIdx].Empty == AUDIOBUF_READY){
//	  Led2_On();
//	  Pause();
  }
//  Led2_Off();

  dac_w = audiobuf[DecodeBufIdx].BaseAddr;
  while (nsamples--) {
		*dac_w = (unsigned short)audio_linear_dither(16, mad_f_mul(*left_ch, VolumeFactor), &dither);
	    dac_w++;
    	*dac_w = *(dac_w - 1);
	    dac_w++;
	    left_ch++;
	    right_ch++;
	    audiobuf[DecodeBufIdx].Size += 4;
  }
  audiobuf[DecodeBufIdx].Empty = AUDIOBUF_READY;
#if defined(USE_I2S_DMA)
  if (DMA_Buf_Empty == 1){
	  PlayingBufIdx = DecodeBufIdx;
	  I2S_DMA_Done = 0;
	  DMA_Buf_Empty = 0;
	  DMA_SendData((uint32_t *)audiobuf[PlayingBufIdx].BaseAddr, audiobuf[PlayingBufIdx].Size);
//	  Led3_Off();
  }
#endif
  DecodeBufIdx++;
  if (DecodeBufIdx == MAX_AUDIOBUF_NUM)
	  DecodeBufIdx = 0;

  return MAD_FLOW_CONTINUE;
}


/*
 * This is the error callback function. It is called whenever a decoding
 * error occurs. The error is indicated by stream->error; the list of
 * possible MAD_ERROR_* errors can be found in the mad.h (or stream.h)
 * header file.
 */
static
enum mad_flow error(void *data,
		    struct mad_stream *stream,
		    struct mad_frame *frame)
{
#if !defined(MP3_16)
  struct buffer *buffer = data;

  fprintf(stderr, "decoding error 0x%04x (%s) at byte offset %u\n",
	  stream->error, mad_stream_errorstr(stream),
	  stream->this_frame - buffer->start);
#endif

  /* return MAD_FLOW_BREAK here to stop decoding (and propagate an error) */

  return MAD_FLOW_CONTINUE;
}


/*
 * This is the function called by main() above to perform all the decoding.
 * It instantiates a decoder object and configures it with the input,
 * output, and error callback functions above. A single call to
 * mad_decoder_run() continues until a callback function returns
 * MAD_FLOW_STOP (to stop decoding) or MAD_FLOW_BREAK (to stop decoding and
 * signal an error).
 * minimad.c(libmad)のdecode関数を元に作成。
 */
static
int decode()
{
//  struct buffer buffer;
  struct mad_decoder decoder;
  int result;

  /* initialize our private message structure */
//  buffer.start  = start;
//  buffer.length = length;

  /* configure input, output, and error functions */
  mad_decoder_init(&decoder, 0 /* &buffer */,
		   input, 0 /* header */, 0 /* filter */, output /*output_monauralize*/,
		   error, 0 /* message */);

  /* start decoding */
  result = mad_decoder_run(&decoder, MAD_DECODER_MODE_SYNC);

  /* release the decoder */
  mad_decoder_finish(&decoder);

  return result;
}

/*
 * MP3ファイルごとの初期処理
 * 	output関数
 * 	I2S設定
 * 	(SysTick設定)
 */
int init_each_mp3(struct mad_decoder *decoder){
	struct mad_header *header;
	header = &(decoder->sync->frame.header);

	//DEBUG_MSG("init_each_mp3 start.\n\r");

	// decoder->output_funcの初期値はoutput_monauralize
    if (header->mode == MAD_MODE_SINGLE_CHANNEL){
    	decoder->output_func = output_mono;
    	ChannelMode = I2S_STEREO;	// 左右同じ音のステレオでI2Sに送信する
    	DEBUG_MSG("init_each_mp3 MONO.\n\r");
    } else if (ChannelMode == 0){		// Speaker Off or Stereo
    	decoder->output_func = output;
    	ChannelMode = I2S_STEREO;
    } else {
    	ChannelMode = I2S_STEREO;
    }
    if (samplerate != header->samplerate /*|| mono != ChannelMode*/){
//    	i2s_init();
    	i2s_stereomono(ChannelMode);
    	mono = ChannelMode;
    	samplerate = header->samplerate;
    	i2s_begin(header->samplerate, 16);
    	DEBUG_MSG("i2s_begin \r\n");
    } else {
//    	i2s_start();
    }
#if !defined(USE_I2S_DMA)
	if (SysTick_Config(SystemCoreClock / header->samplerate - 1)) {
	}
#endif
	return MAD_FLOW_CONTINUE;
}

#if defined(USE_I2S_DMA)

/*----------------------------------------------------------------------------
  GPDMA functions
 *----------------------------------------------------------------------------*/

/**
  * @brief  Initializes GPDMA.
  *
  * @param  None
  * @retval None
  *
  */
void DMA_Init(void)
{
	/* Enable GPDMA clock */
	LPC_SC->PCONP |= (1 << 29);
	/* init DMA	channel x */
    LPC_GPDMA -> DMACIntTCClear = (1 << I2S_DMA_CHAN);
    LPC_GPDMA -> DMACIntErrClr = (1 << I2S_DMA_CHAN);

	LPC_GPDMACH->DMACCDestAddr = DMA_I2S_TX_FIFO;
	LPC_GPDMACH -> DMACCLLI = 0;
	LPC_GPDMACH->DMACCControl = (0x04 << 12) | (0x04 << 15)
	  	| (0x02 << 18) | (0x02 << 21)
		| (1 << 26) | 0x80000000;
	LPC_GPDMACH->DMACCConfig = 0x18000 | (0x00 << 1) | (0x05 << 6) | (0x01 << 11);

    /* Enable DMA channels, little endian */
    LPC_GPDMA->DMACConfig = 0x01;
    while ( !(LPC_GPDMA->DMACConfig & 0x01) );

	/* Enalbe I2S DMA TX */
	LPC_I2S->I2SDMA1 = (0x01<<1) | (0x04<<16);

  	NVIC_EnableIRQ(DMA_IRQn);
	NVIC_SetPriority (DMA_IRQn, 1);
}

/**
  * @brief  Start GPDMA to transer data from memory to peripheral.
  *
  * @param  buf: Pointer to the buffer
  * @param  size: data length in byte
  * @retval None
  *
  */
void DMA_SendData(uint32_t *buf, uint32_t size)
{
	LPC_GPDMACH->DMACCSrcAddr = (uint32_t)buf;
	LPC_GPDMACH->DMACCControl |= (size/4) & 0x0FFF; /* burst size=4 bytes */
	LPC_GPDMACH->DMACCConfig |= 0x01; /* Start DMA to send data */
}

/**
  * @brief  DMA ISR.
  *
  * @param  None
  * @retval None
  *
  */
void DMA_IRQHandler(void)
{
    uint32_t regVal;

//	DEBUG_MSG("DMA_IRQHandler\n\r");
    regVal = LPC_GPDMA->DMACIntErrStat;
    if ( regVal )
    {
        LPC_GPDMA->DMACIntErrClr |= regVal;
    }

    regVal = LPC_GPDMA->DMACIntTCStat;
    if ( regVal )
    {
        LPC_GPDMA->DMACIntTCClear |= regVal;

	    if (regVal & (1<<I2S_DMA_CHAN))
	    {
		    I2S_DMA_Done = 1;

//            frame_out_flag = !frame_out_flag;
//            if (frame_out_flag) LED_On (LED_FRAME_OUT);
//            else                LED_Off (LED_FRAME_OUT);

            /* Data in audiobuf[PlayingBufIdx] has been sent out */
    		audiobuf[PlayingBufIdx].Empty = AUDIOBUF_EMPTY;
    		audiobuf[PlayingBufIdx].Size = 0;

		    /* Send the data in next audio buffer */
		    PlayingBufIdx++;
		    if (PlayingBufIdx == MAX_AUDIOBUF_NUM)
			    PlayingBufIdx = 0;
    	#if 1
    		if (audiobuf[PlayingBufIdx].Empty == AUDIOBUF_EMPTY)		// デコードが間に合ってない
    		{
                /* If empty==1, it means read file+decoder is slower than playback
                (it will cause noise) or playback is over (it is ok). */
//    			DEBUG_MSG("\n\r**Play buf(%d) is empty!\n\r", PlayingBufIdx);
//    			decode_slower_flag ^= 0x1;
//    			if (decode_slower_flag) LED_On (LED_DEC_SLOWER);
//    			else                    LED_Off (LED_DEC_SLOWER);

                DMA_Buf_Empty = 1;
                Led3_On();
                return;
            } else {
            	Led3_Off();
            }
    	#endif

		    DMA_SendData((uint32_t *)audiobuf[PlayingBufIdx].BaseAddr, audiobuf[PlayingBufIdx].Size);
		    I2S_DMA_Done = 0;
	    }
    }
}

#else		// !defined(USE_I2S_DMA)
/*
 * SysTickタイマー　オフ
 */
void SysTick_Off(){
	SysTick->CTRL &= ~SysTick_CTRL_ENABLE_Msk;
}

/*
 * SysTickタイマー　オン
 */
void SysTick_On(){
	SysTick->CTRL |= SysTick_CTRL_ENABLE_Msk;
}

//  SysTick
void SysTick_Handler(void) {
	if(dac_s < dac_e){
		i2s_write(dac_s, 2);
		dac_s += 2;
	}
	if(dac_s >= dac_e){
		uint8_t next_playing_buf_idx;

        /* Data in audiobuf[PlayingBufIdx] has been sent out */
		audiobuf[PlayingBufIdx].Empty = AUDIOBUF_EMPTY;
		audiobuf[PlayingBufIdx].Size = 0;

		next_playing_buf_idx = PlayingBufIdx + 1;
	    if (next_playing_buf_idx == MAX_AUDIOBUF_NUM)
	    	next_playing_buf_idx = 0;
		if (audiobuf[next_playing_buf_idx].Empty == AUDIOBUF_READY){
			PlayingBufIdx = next_playing_buf_idx;
		    dac_s = (dacout_s *)(audiobuf[PlayingBufIdx].BaseAddr);
		    dac_e = dac_s + (audiobuf[PlayingBufIdx].Size / sizeof(dacout_s));
			Led3_Off();
		} else {	// デコードが間にあってない
			Led3_On();
		}
	}
}
#endif		// defined(USE_I2S_DMA)

// TIMER0
void TIMER0_IRQHandler (void)
{
	LPC_TIM0->IR = 1;			/* clear interrupt flag */
	timer0_counter++;

	return;
}

void Pause(){
	timer0_counter = 0;
	while (timer0_counter == 0){

	}
}

// TIMER1
void TIMER1_IRQHandler (void){

	LPC_TIM1->IR = 1;			/* clear interrupt flag */
	timer1_counter++;
	if (timer1_counter >= TIMER1_1MIN){
		timer1_counter = 0;
		AutoPoffTimer++;		// １分ごとにインクリメント
	}

    if (CHATTERING_TMR == 0){

    	if ((LPC_GPIO2->FIOPIN & (1 << 5)) == 0) {					// P2.5
    		PinEvent = SWSELECT_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN & (1 << 4)) == 0) {			// P2.4
			PinEvent = SWRETURN_PRESSED;
		} else if ((LPC_GPIO2->FIOPIN & (1 << 3)) == 0){			// P2.3
			if (NowPlaying == 0){
				PinEvent = SWDOWN_PRESSED;
			} else {
				if (VolumeIdx > 0){
					VolumeIdx--;
					VolumeFactor = mad_f_tofixed((long)VolumeTab[VolumeIdx]/(double)MAX_VOLUME);
				}
				CHATTERING_TMR = 1;
			}
    	} else if ((LPC_GPIO2->FIOPIN & (1 << 2)) == 0){			// P2.2
			if (NowPlaying == 0){
				PinEvent = SWUP_PRESSED;
			} else {	// 再生中は音量調整 (この関数内で処理してしまうのでPinEventを設定しない)
				if (VolumeIdx < MAX_VOLUMETAB_IDX){
					VolumeIdx++;
					VolumeFactor = mad_f_tofixed((long)VolumeTab[VolumeIdx]/(double)MAX_VOLUME);
				}
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO2->FIOPIN & (1 << 1)) == 0) {	// P2.1
			PinEvent = SWOPTION_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN & (1 << 11)) == 0) {	// P0.11
			PinEvent = SWPLAY_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN & (1 << 10)) == 0) {	// P0.10
			PinEvent = SWRANDOM_PRESSED;
		} else if ((LPC_GPIO0->FIOPIN0 & (1 << 5)) == 0){	// P0.5
			if (NowPlaying == 0){
				PinEvent = SWPAGEDOWN_PRESSED;
			} else {	// 再生中は音量調整 (この関数内で処理してしまうのでPinEventを設定しない)
				if (VolumeIdx > VOLUMEIDX_STEP){
					VolumeIdx -= VOLUMEIDX_STEP;
					VolumeFactor = mad_f_tofixed((long)VolumeTab[VolumeIdx]/(double)MAX_VOLUME);
				}
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO0->FIOPIN0 & (1 << 4)) == 0){	// P0.4
			if (NowPlaying == 0){
				PinEvent = SWPAGEUP_PRESSED;
			} else {	// 再生中は音量調整 (この関数内で処理してしまうのでPinEventを設定しない)
				if (VolumeIdx < (MAX_VOLUMETAB_IDX - VOLUMEIDX_STEP)){
					VolumeIdx += VOLUMEIDX_STEP;
					VolumeFactor = mad_f_tofixed((long)VolumeTab[VolumeIdx]/(double)MAX_VOLUME);
				}
				CHATTERING_TMR = 1;
			}
		} else if ((LPC_GPIO1->FIOPIN & (1 << 31)) == 0){	// P1.31
			PinEvent = SWPOFF_PRESSED;
		} else {
			PinEvent = 0;
		}

		if (PinEvent != 0){
			CHATTERING_TMR = 1;
		}
    } else if (CHATTERING_TMR > 0){
    	CHATTERING_TMR++;
    	if (CHATTERING_TMR == 2){		// 2020/12/09 3->2
    		CHATTERING_TMR = 0;
    	}
    }

}


// ミリ秒単位の待ち時間
// TIMER0を使うので再生中は使えない。
void WaitMs(int ms){
	ms *= 10;
	timer0_counter = 0;
	while (timer0_counter < ms){

	}
}

#if 0
void test_read_write(){
//	FIL 		Fil1;
//	FIL 		Fil2;
	FIL 		*pFil1;
	FIL 		*pFil2;
    FRESULT     res;
    UINT		br;
    UINT		br2;
    char		buff[1024];

    pFil1 = (FIL *)mad_malloc(sizeof(FIL));
    pFil2 = (FIL *)mad_malloc(sizeof(FIL));

    res = f_open(pFil1 /*&Fil1*/, "/MSC1_in/test.dat", FA_READ | FA_OPEN_EXISTING);
    if (res != FR_OK){
//        DEBUG_MSG ("f_open failed.\n\r");
        return ;
    }
    res = f_open(pFil2 /*&Fil2*/, "/MSC1_out/testout.dat", FA_WRITE | FA_CREATE_ALWAYS);
    if (res != FR_OK){
//        DEBUG_MSG ("f_open failed.\n\r");
        return ;
    }
    do {
    	res = f_read(pFil1 /*&Fil1*/, buff, 1024, &br);
    	if (res != FR_OK || br == 0){
    		break;
    	}
    	//DEBUG_MSG(buff);
    	res = f_write(pFil2 /*&Fil2*/, buff, br, &br2);
		if (res != FR_OK){
			break;
		}
    } while (1);

    f_close(pFil1 /*&Fil1*/);
    f_close(pFil2 /*&Fil2*/);
//    DEBUG_MSG ("test_read_write OK.\n\r");
}

#endif

// 電源遮断
void PowerOff(){
	SpeakerAmpDisable();
	LPC_GPIO1->FIOCLR |= (0x01 << POWER_ON_BIT);		// P1.30 (set low)
    while(1) {
    }
}

/*
 * スピーカーアンプをActiveに
 */
void SpeakerAmpEnable(){
    LPC_GPIO0->FIOSET |= (0x01 << SPEAKER_MUTE_BIT);
}

/*
 * スピーカーアンプをInactiveに
 */
void SpeakerAmpDisable(){
	LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_MUTE_BIT);
}

/*
 * 出力先：スピーカORイヤホン切り替え（リレー駆動）
 */
void SetOutput(){
	DebugWritelnV1("SetOutput=", RecHistory[HS_OUTPUT_OFFSET]);
	if (RecHistory[HS_OUTPUT_OFFSET] == OUTPUT_SPEAKER){
	    LPC_GPIO0->FIOSET |= (0x01 << SPEAKER_OR_EARPHONE);
	} else {
		LPC_GPIO0->FIOCLR |= (0x01 << SPEAKER_OR_EARPHONE);
	}
}

/*
 * 出力先がスピーカのときスピーカーアンプのイネーブル切り替え
 * onoff=1のときスピーカーアンプ・イネーブル
 */
void ActivateSpeaker(int onoff){
	if (RecHistory[HS_OUTPUT_OFFSET] == OUTPUT_SPEAKER){
		if (onoff == 1){
			SpeakerAmpEnable();
		} else {
			SpeakerAmpDisable();
		}
	}
}


/*
 * ディレクトリの選択操作
 * 	レコード番号：ヘッダレコードを０とする。以降、ルートディレクトリを１、など。
 * 	RecBufはDIR構成ファイルまたはMP3ファイルリストの１レコードを取得できるサイズを持つ。
 * 	FilDL:DIR構成ファイル、FilML:MP3ファイルリスト、いずれもf_open済み。Fil3はその他任意のファイルでそのつどf_open/fcloseする。
 */
int SelectDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3){
    FRESULT     res;
    uint32_t    ofsSubDirStart;
    uint16_t    rowSubDirStart;			// 現在対象のディレクトリの先頭サブディレクトリのレコード番号
    uint16_t    numSubDir;				// サブディレクトリ数
    uint16_t    idxSubDir;				// 選択ページ先頭に表示中のDIR (0,1,..., numSubDir-1)
    uint8_t     idxCurSubDirItem;   	// 現在表示ページで選択中のDIR(0,1,2,3)
    uint8_t     idxCurSubDirItemBk;		// 上記の一時的退避
    uint8_t		NumSelItem;				// １ページに表示する最大項目数
    int        	ret;
    uint8_t     idxDirDepth;
    uint16_t    stackIdxSubDir[8];
    uint8_t     stackIdxCurrentSubDirItem[8];
    uint16_t    dirno[MAX_SELDIR_ITEM];
    uint16_t    numSubDirTab[MAX_SELDIR_ITEM];
    uint16_t    numSndFile[MAX_SELDIR_ITEM];
    uint16_t    recNoFirstSubdir[MAX_SELDIR_ITEM];
    uint8_t     readdlflg;      // DIR LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;
    UINT    	br;

    DEBUG_MSG("SelectDirectory start\r\n");
    ret = FR_OK;

    // Skip HEADER info
    res = f_lseek(FilDL, RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_lseek error\n\r");
        return res;
    }

    // Read ROOT info
    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST, &br);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_read error (root)\r\n");
        return res;
    }

    // ROOTディレクトリの最初のサブディレクトリのレコード番号
	memcpy(&rowSubDirStart, RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
	//DebugWritelnV1("f_read_DIRLISTFILE 2 rowSubDirStart ", rowSubDirStart);
    ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
    numSubDir = *(RecBuf + DL_NUM_SUBDIR_OFFSET);
    idxSubDir = 0;
    idxCurSubDirItem = 0;
    idxCurSubDirItemBk = -1;
    idxDirDepth = 0;
    readdlflg = TRUE;
    NumSelItem = MAX_SELDIR_ITEM;

    // end point of ROOT info
    res = f_lseek(FilDL, 2*RECLEN_DIRLIST_FULL);		// ヘッダレコードとルートディレクトリレコードを読み飛ばす
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_lseek error (end of root)\r\n");
        return res;
    }

    while (1){
    	if (readdlflg == TRUE){
    		items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			res = f_lseek(FilDL, ofsSubDirStart + (uint32_t)idxSubDir * RECLEN_DIRLIST_FULL);
			for (i=0; i<NumSelItem; i++){
				if ((idxSubDir + i) < numSubDir){
					if ((res = f_read(FilDL, RecBuf, RECLEN_DIRLIST_FULL, &br)) == FR_OK){

						//dirno[i] = rowSubDirStart + idxSubDir + i - 1;	// MP3_12のバグ？
						dirno[i] = rowSubDirStart + idxSubDir + i;
                        numSubDirTab[i] = *(RecBuf + DL_NUM_SUBDIR_OFFSET);
						numSndFile[i] = *(RecBuf + DL_NUM_SNDFILE_OFFSET);
						memcpy(&recNoFirstSubdir[i], RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
						items++;
						// イメージ
						memcpy(LcdImageBuff + 2*i*OLED_WIDTH, RecBuf + RECLEN_DIRLIST, OLED_WIDTH);
						memcpy(LcdImageBuff + 2*i*OLED_WIDTH + OLED_WIDTH, RecBuf + RECLEN_DIRLIST + DL_IMAGEDATAHALF_LEN, OLED_WIDTH);
					} else {
						// エラー
				    	DispMsg(i, MSGNO_DIRLST_ERR, 0, Fil3);
					}
				} else {
					dirno[i] = 0;
					numSndFile[i] = 0;
					recNoFirstSubdir[i] = 0;
				}
			}
			SetMenuItemSelected(idxCurSubDirItem);
        	DispLcdFullImage();
    	}

		//DebugWritelnV1("idxCurrentSubDirItem=", idxCurrentSubDirItem);
		ret = PauseAndSelectItem();
	    readdlflg = TRUE;
		//DebugWritelnV1("PauseAndSelectItem=", ret);
		if (ret == SWUP_PRESSED){
			// 同一ページ内の上の行へ、または、前ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == 0){
				if (idxSubDir >= NumSelItem){
					idxSubDir -= NumSelItem;	// 前ページへ
					idxCurSubDirItem = NumSelItem - 1;
				} else {
					readdlflg = FALSE;
				}
			} else {
				idxCurSubDirItem--;
				readdlflg = FALSE;
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk);
				SetMenuItemSelected(idxCurSubDirItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){
			// 同一ページ内の下の行へ、または、次ページへ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxCurSubDirItem % LCD_STRING_LINES) == (LCD_STRING_LINES - 1)){
				if ((idxSubDir + NumSelItem) < numSubDir){
					idxSubDir += NumSelItem;	// 次のページへ
					idxCurSubDirItem = 0;
				} else {
					readdlflg = FALSE;
				}
			} else {
				if (idxCurSubDirItem < (items - 1)){
					idxCurSubDirItem++;
					readdlflg = FALSE;
				} else {
					readdlflg = FALSE;
				}
			}
			if (readdlflg == FALSE && idxCurSubDirItem != idxCurSubDirItemBk){
				SetMenuItemSelected(idxCurSubDirItemBk);
				SetMenuItemSelected(idxCurSubDirItem);
	        	DispLcdFullImage();
			}
		} else if (ret == SWPAGEUP_PRESSED){		// 改ページ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if (idxSubDir >= NumSelItem){
				idxSubDir -= NumSelItem;	// 前ページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPAGEDOWN_PRESSED){	// 改ページ
			idxCurSubDirItemBk = idxCurSubDirItem;
			if ((idxSubDir + NumSelItem) < numSubDir){
				idxSubDir += NumSelItem;	// 次のページへ
				idxCurSubDirItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPLAY_PRESSED){		// PLAY
			if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				ActivateSpeaker(1);
				ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, dirno[idxCurSubDirItem], 0, 0, PD_SEQUENTIAL);
				ActivateSpeaker(0);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					//break;
				}
			} else if (idxDirDepth == 0) {	// ROOTディレクトリでは前回再生の再開（ROOTディレクトリにはMP3ファイルを置かない）
				uint16_t	last_dirno;
				uint8_t		last_trackno;
				uint32_t	last_bytes;

				memcpy(&last_dirno, &RecHistory[HS_LASTDIR_OFFSET], HS_LASTDIR_LEN);
				last_trackno = RecHistory[HS_LASTTRACK_OFFSET];
				memcpy(&last_bytes, &RecHistory[HS_LASTBYTE_OFFSET], HS_LASTBYTE_LEN);
				if (last_dirno > 1){	// 前回再生の履歴が存在すること
					ActivateSpeaker(1);
					ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, last_dirno, last_trackno, last_bytes, PD_SEQUENTIAL);
					ActivateSpeaker(0);
					if (ret == POWEROFF){
						break;
					} else if (ret == GOTO_MAINMENU){
						//break;
					}
				}
			} else {
				//
		    	DEBUG_MSG("No mp3 files\n\r");
			}
		} else if (ret == SWRANDOM_PRESSED){	// RANDOM
			ret = START_RANDOM;
			break;
		} else if (ret == SWSELECT_PRESSED){  // 下位DIRへ (SKIP forward directory)
			if (recNoFirstSubdir[idxCurSubDirItem] > 1){		// 下位DIRが存在する
				rowSubDirStart = recNoFirstSubdir[idxCurSubDirItem];
				DebugWritelnV1(" rowSubDirStart=", rowSubDirStart);
				ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
				numSubDir = numSubDirTab[idxCurSubDirItem];		// *(RecDirList + DL_NUM_SUBDIR);
				// push current idxSubDir to stack
				stackIdxSubDir[idxDirDepth] = idxSubDir;
				stackIdxCurrentSubDirItem[idxDirDepth] = idxCurSubDirItem;
				idxDirDepth++;
				idxSubDir = 0;
				idxCurSubDirItem = 0;
				//DebugWritelnV1("rowSubDirStart=", rowSubDirStart);
		        //NumSelItem = MAX_SELDIR_ITEM;		//初期値に戻さない
			} else if (numSndFile[idxCurSubDirItem] != 0){      // Have a MP3 file ?
				res = f_lseek(FilDL, ofsSubDirStart + (uint32_t)(idxSubDir + idxCurSubDirItem) * RECLEN_DIRLIST_FULL);
			    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST, &br);

				ret = SelectSoundfile(RecBuf, FilDL, FilML, Fil3, dirno[idxCurSubDirItem], 0);
				if (ret == POWEROFF){
					break;
				} else if (ret == GOTO_MAINMENU){
					//break;
				}
			}
		} else if (ret == SWRETURN_PRESSED){  // 上位DIRへ (GOTO Mainmenu)
			uint16_t    r;
			uint16_t    r2;

			//r = *((uint16_t *)(RecDirList + DL_RECNO_GRANDPARENTDIR));   // rec no of grand parent
			//r = P8U16(RecDirList, DL_RECNO_GRANDPARENTDIR);		// rec no of grand parent
			// RecDirListは表示中のディレクトリのどれかのDIR構成ファイルレコードを保持していることが前提！
			memcpy(&r,  RecBuf + DL_RECNO_GRANDPARENTDIR_OFFSET, DL_RECNO_GRANDPARENTDIR_LEN);	// rec no of grand parent
			memcpy(&r2, RecBuf + DL_RECNO_PARENTDIR_OFFSET, DL_RECNO_PARENTDIR_LEN);
			// 親々ディレクトリのレコード番号と親ディレクトリのレコード番号が一致しているときは、トップレベルのDIR階層
			if (r == r2){
				//なにもしない
				//ret = GOTO_MAINMENU;
				//break;
			} else {
				// Go to grand parent record
				res = f_lseek(FilDL, (uint32_t)r * RECLEN_DIRLIST_FULL);
				if (res != FR_OK) return res;
				// Read grand parent record
				//res = f_read_DIRLISTFILE(Fil, RecDirList);
			    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST, &br);
				if (res != FR_OK) return res;
				// the first "uncle叔父" record
				//r = *((uint16_t *)(RecDirList + DL_RECNO_SUBDIR_START));
				//r = P8U16(RecDirList, DL_RECNO_SUBDIR_START);
				memcpy(&r, RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
				if (r > 1){
					rowSubDirStart = r;
					ofsSubDirStart = (uint32_t)rowSubDirStart * RECLEN_DIRLIST_FULL;
					numSubDir = *(RecBuf + DL_NUM_SUBDIR_OFFSET);
					// pop idxSubDir from stack -> This is parent
					idxDirDepth--;
					idxSubDir = stackIdxSubDir[idxDirDepth];
					idxCurSubDirItem = stackIdxCurrentSubDirItem[idxDirDepth];
					if (idxSubDir >= numSubDir) idxSubDir = 0;
					readdlflg = TRUE;
				} else {
					//なにもしない
					//ret = GOTO_MAINMENU;
					//break;
				}
			}
			//readdlflg = TRUE;
		} else if (ret == SWOPTION_PRESSED){      //
			// 出力先変更：スピーカORイヤホン
			SelectOption(Fil3);
		    SetOutput();
			readdlflg = TRUE;
		} else if (ret == SWPOFF_PRESSED){
			ret = POWEROFF;
			break;
		}

    }

    DEBUG_MSG("SelectDirectory end\r\n");
    return ret;
}

/*
 * 音源ファイルの選択操作
 * 	FilDL:DIR構成ファイル、FilML:MP3ファイルリスト、いずれもf_open済み。Fil3はその他任意のファイルでそのつどf_open/fcloseする。
 */
int SelectSoundfile(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno){
    FRESULT     res;
    int		    ret;
    //uint16_t    rowSubDirStart;	// 対象ディレクトリの先頭サブディレクトリのレコード番号
    uint16_t 	numsndfile;		// 対象ディレクトリのMP3ファイル数
    uint16_t	rowSndStart;	// 対象ディレクトリのMP3ファイルリスト先頭レコード番号
    uint8_t     trackno;		// 選択中の曲番 (0,1,2,...,numsndfile-1)
    uint8_t     idxtrack;       // 選択操作中のページ先頭の曲番 (0,1,2,...,numsndfile-1)
    uint8_t		NumSelItem;		// １ページに表示する項目数
    UINT        br;
    uint16_t    dirbytes;
    uint8_t     readdlflg;      // MP3 LIST ファイルを読む必要があるか？
    uint8_t		i;
    uint8_t		items = 0;
    uint16_t    idxCurTitleItem;
    uint16_t    idxCurTitleItemBk;

    ret = FR_OK;

	DEBUG_MSG("SelectSoundfile start\n\r");
	DebugWritelnV1(" dirno=", dirno);
	res = f_lseek(FilDL, (uint32_t)(dirno) * RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_lseek error\n\r");
        return res;
    }
    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST_FULL, &br);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_read error\n\r");
        return res;
    }
    memcpy(&dirbytes, RecBuf + DL_FULLDIRBYTES_OFFSET, DL_FULLDIRBYTES_LEN);
    numsndfile = RecBuf[DL_NUM_SNDFILE_OFFSET];
    if (numsndfile == 0){
    	DEBUG_MSG("no mp3 files\n\r");
    	return FR_OK;
    }
	memcpy(&rowSndStart, RecBuf + DL_RECNO_SNDFILE_OFFSET, DL_RECNO_SNDFILE_LEN);
	//memcpy(&rowSubDirStart, RecBuf + DL_RECNO_SUBDIR_START_OFFSET, DL_RECNO_SUBDIR_START_LEN);
//	DebugWriteln((char *)(RecDirList + DL_FULLDIR));

    if (start_trackno >= numsndfile) start_trackno = 0;
    trackno = start_trackno;
    idxtrack = 0;
    idxCurTitleItem = 0;
    readdlflg = TRUE;
    NumSelItem = MAX_SELMP3_ITEM;

    while (1){
		if (readdlflg == TRUE){		//
			// 選択肢を表示
			items = 0;
			memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
			res = f_lseek(FilML, (uint32_t)(rowSndStart + idxtrack) * RECLEN_MP3LIST_FULL);
			for (i=0; i<NumSelItem; i++){
				if ((idxtrack + i) < numsndfile){
					if ((res = f_read(FilML, RecBuf, RECLEN_MP3LIST_FULL, &br)) == FR_OK){
						items++;
						// イメージ
						memcpy(LcdImageBuff + 2*i*OLED_WIDTH, RecBuf + RECLEN_MP3LIST, OLED_WIDTH);
						memcpy(LcdImageBuff + 2*i*OLED_WIDTH + OLED_WIDTH, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN, OLED_WIDTH);
					} else {
						// エラー
						DispMsg(i, MSGNO_SNDLST_ERR, 0, Fil3);
					}
				}
			}
			SetMenuItemSelected(idxCurTitleItem);
			DispLcdFullImage();
		}

		ret = PauseAndSelectItem();
		readdlflg = TRUE;
		if (ret == SWUP_PRESSED){
			idxCurTitleItemBk = idxCurTitleItem;
			if (idxCurTitleItem == 0){
				//idxCurTitleItem = NumSelItem - 1;
				if (idxtrack >= NumSelItem){
					idxtrack -= NumSelItem;
					//idxCurTitleItem = 0;
					idxCurTitleItem = NumSelItem - 1;
				} else {
					readdlflg = FALSE;
				}
			} else {
				idxCurTitleItem--;
				readdlflg = FALSE;
			}
			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
				SetMenuItemSelected(idxCurTitleItemBk);
				SetMenuItemSelected(idxCurTitleItem);
				DispLcdFullImage();
			}
		} else if (ret == SWDOWN_PRESSED){
			idxCurTitleItemBk = idxCurTitleItem;
			if (idxCurTitleItem < (NumSelItem - 1)){
				if ((idxtrack + idxCurTitleItem) < (numsndfile - 1)){
					idxCurTitleItem++;
					readdlflg = FALSE;
				} else {
					readdlflg = FALSE;
				}
			} else {
				//idxCurTitleItem = 0;
				if ((idxtrack + NumSelItem) < numsndfile){
					idxtrack += NumSelItem;
					idxCurTitleItem = 0;
				} else {
					readdlflg = FALSE;
				}
			}
			if (readdlflg == FALSE && idxCurTitleItem != idxCurTitleItemBk){
				SetMenuItemSelected(idxCurTitleItemBk);
				SetMenuItemSelected(idxCurTitleItem);
				DispLcdFullImage();
			}
		} else if (ret == SWPAGEUP_PRESSED){		// 改ページ
			if (idxtrack >= NumSelItem){
				idxtrack -= NumSelItem;
				idxCurTitleItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPAGEDOWN_PRESSED){	// 改ページ
			if ((idxtrack + NumSelItem) < numsndfile){
				idxtrack += NumSelItem;
				idxCurTitleItem = 0;
			} else {
				readdlflg = FALSE;
			}
		} else if (ret == SWPLAY_PRESSED){
			//DebugWriteln("SWPLAY_PRESSED pressed \r\n");
			trackno = idxtrack + idxCurTitleItem;
			ActivateSpeaker(1);
			ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, dirno, trackno, 0, PD_SEQUENTIAL);
			ActivateSpeaker(0);
			if (ret == POWEROFF){
				break;
			} else if (ret == GOTO_MAINMENU){
				break;
			}
		} else if (ret == SWRETURN_PRESSED){
			ret = PLAY_OTHER_DIR;
			break;			// 現ディレクトリを抜ける。
		} else if (ret == SWOPTION_PRESSED){
			// 出力先変更：スピーカORイヤホン
			SelectOption(Fil3);
			SetOutput();
			readdlflg = TRUE;
		} else if (ret == SWPOFF_PRESSED){
			ret = POWEROFF;
			break;
		}

    }

	DEBUG_MSG("SelectSoundfile end\n\r");
    return ret;
}

/*
 * ディレクトリ下のMP3ファイルを順次再生
 * 	FilDL:DIR構成ファイル、FilML:MP3ファイルリスト、いずれもf_open済み。Fil3はその他任意のファイルでそのつどf_open/fcloseする。
 */
int PlayOneDirectory(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3, uint16_t dirno, uint8_t start_trackno, uint32_t start_byte, playdirectory_t mode){
    FRESULT     res;
    int		    ret;
    uint8_t		filefullpath[DL_FULLDIR_LEN + ML_FILENAME_LEN + 2];
    uint16_t 	numsndfile;		// 対象ディレクトリのMP3ファイル数
    uint16_t	rowSndStart;	// 対象ディレクトリのMP3ファイルリスト先頭レコード番号
    uint8_t     trackno;		// 再生中の曲番 (0,1,2,...,numsndfile-1)
    UINT        br;
    uint16_t    dirbytes;
    uint8_t     execplayflg = 0;
    int			vflen;

    ret = FR_OK;

	DEBUG_MSG("SelectSoundfile start\n\r");
	DebugWritelnV1(" dirno=", dirno);
	res = f_lseek(FilDL, (uint32_t)(dirno) * RECLEN_DIRLIST_FULL);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_lseek error\n\r");
        return res;
    }
    res = f_read(FilDL, RecBuf, RECLEN_DIRLIST_FULL, &br);
    if (res != FR_OK){
    	DispMsg(0, MSGNO_DIRLST_ERR, 0, Fil3);
    	WaitMs(3000);
    	DEBUG_MSG("f_read error\n\r");
        return res;
    }
    memcpy(filefullpath, RecBuf + DL_FULLDIR_OFFSET, DL_FULLDIR_LEN);
    memcpy(&dirbytes, RecBuf + DL_FULLDIRBYTES_OFFSET, DL_FULLDIRBYTES_LEN);
    numsndfile = RecBuf[DL_NUM_SNDFILE_OFFSET];
    if (numsndfile == 0){
    	DEBUG_MSG("no mp3 files\n\r");
    	return FR_OK;
    }
	memcpy(&rowSndStart, RecBuf + DL_RECNO_SNDFILE_OFFSET, DL_RECNO_SNDFILE_LEN);
	// １行目にディレクトリ名
	memcpy(LcdImageBuff, RecBuf + RECLEN_DIRLIST, OLED_WIDTH);
	memcpy(LcdImageBuff + OLED_WIDTH, RecBuf + RECLEN_DIRLIST + DL_IMAGEDATAHALF_LEN, OLED_WIDTH);
    // ディレクトリパスとファイル名との間の区切り (UNICODEの'\')
    filefullpath[dirbytes++] = '\\';
    filefullpath[dirbytes++] = 0;
//	DebugWriteln((char *)(RecDirList + DL_FULLDIR));
    if (start_trackno >= numsndfile) start_trackno = 0;

    // ボリュームファイルによる音量設定
    vflen = 2*strlen(VOLUMEFILE);		// UNICODE長
    memcpy(filefullpath + dirbytes, MakeUnicodeString(VOLUMEFILE), vflen);
    memset(filefullpath + dirbytes + vflen, 0, 2);
    SetLocalVolume((TCHAR *)filefullpath, RecHistory[HS_OUTPUT_OFFSET], Fil3);

    for (trackno=start_trackno; trackno<numsndfile; trackno++){
    	if (mode == PD_ONLY_ONE && trackno != start_trackno) continue;
        if (f_lseek(FilML, (uint32_t)(rowSndStart + trackno)*(RECLEN_MP3LIST_FULL)) != FR_OK) break;
        if (f_read(FilML, RecBuf, RECLEN_MP3LIST_FULL, &br) != FR_OK) break;
		// ２，３，４行目に曲名表示
		memcpy(LcdImageBuff + 2*1*OLED_WIDTH, RecBuf + RECLEN_MP3LIST, OLED_WIDTH);
		memcpy(LcdImageBuff + 2*2*OLED_WIDTH, RecBuf + RECLEN_MP3LIST + OLED_WIDTH, OLED_WIDTH);
		memcpy(LcdImageBuff + 2*3*OLED_WIDTH, RecBuf + RECLEN_MP3LIST + 2*OLED_WIDTH, OLED_WIDTH);
		memcpy(LcdImageBuff + (2*1 + 1)*OLED_WIDTH, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN, OLED_WIDTH);
		memcpy(LcdImageBuff + (2*2 + 1)*OLED_WIDTH, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN + OLED_WIDTH, OLED_WIDTH);
		memcpy(LcdImageBuff + (2*3 + 1)*OLED_WIDTH, RecBuf + RECLEN_MP3LIST + ML_IMAGEDATAHALF_LEN + 2*OLED_WIDTH, OLED_WIDTH);
    	DispLcdFullImage();

        // Mp3 file full path
        memcpy(filefullpath + dirbytes, RecBuf + ML_FILENAME_OFFSET, ML_FILENAME_LEN);
        // Play a MP3 file
        ret = PlayOneMp3((TCHAR *)filefullpath, dirno, trackno, start_byte);
        //DebugWritelnV1("PlayOneMp3=", ret);
        if (ret == PLAY_NEXT_MP3){				// SWPLAY
            // 再生続行
        } else if (ret == PLAY_OTHER_DIR){		// SWRANDOM（再生中にRANDOMボタンが使えるのはランダム再生中だけ）
        	break;
        } else if (ret == POWEROFF){			// SWPOWER
            break;
        } else if (ret == SELECT_OTHER_MP3){	// SWRETURN
        	ret = GOTO_MAINMENU;
        	break;
        } else {
            execplayflg = 1;
        }
        start_byte = 0;
    }

    // ボリュームファイル保存
    if (execplayflg == 1){
        memcpy(filefullpath + dirbytes, MakeUnicodeString(VOLUMEFILE), vflen);
        memset(filefullpath + dirbytes + vflen, 0, 2);
        SaveVolumeFile((TCHAR *)filefullpath, RecHistory[HS_OUTPUT_OFFSET], Fil3);
    }
    // 履歴保存
    SaveHistory(Fil3);

    return ret;
}


/*
 * ランダム再生
 */
int PlayRandom(uint8_t *RecBuf, FIL *FilDL, FIL *FilML, FIL *Fil3){
    FRESULT     res;
    int		    ret;
    FIL 		FilRL;
    uint8_t		rec_random[RECLEN_RANDOMLIST];
    uint8_t		unit;
    uint16_t	recno_random;
    uint16_t	recno_dir;
    uint8_t		trackno;
    UINT		bytes;

    ret = FR_OK;

	res = f_open(&FilRL, MakeUnicodeString(RANDOMORDERFILE), FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DispMsg(0, MSGNO_RANDOM_ERR, 0, Fil3);
		WaitMs(2000);
		return FALSE;
	}
	memcpy(&recno_random, RecHistory + HS_RANDOMREC_OFFSET,  HS_RANDOMREC_LEN);
    res = f_lseek(&FilRL, (DWORD)recno_random*RECLEN_RANDOMLIST);
	if (res != FR_OK){
	    res = f_lseek(&FilRL, 0);
	}
	ActivateSpeaker(1);
	RandomPlaying = 1;

	do {
		res = f_read(&FilRL, rec_random, RECLEN_RANDOMLIST, &bytes);
		if (res != FR_OK || bytes != RECLEN_RANDOMLIST){
		    res = f_lseek(&FilRL, 0);
			if (res == FR_OK){
				recno_random = 0;
				continue;
			} else {
				break;
			}
		}
		unit = rec_random[RD_UNIT_OFFSET];
		memcpy(&recno_dir, &rec_random[RD_DIRRECNO_OFFSET], RD_DIRRECNO_LEN);
		trackno = rec_random[RD_TRACKNO_OFFSET];

		recno_random++;
		memcpy(RecHistory + HS_RANDOMREC_OFFSET, &recno_random, HS_RANDOMREC_LEN);

		if (unit == PD_ONLY_ONE){
			ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, recno_dir, trackno, 0, PD_ONLY_ONE);
		} else {
			ret = PlayOneDirectory(RecBuf, FilDL, FilML, Fil3, recno_dir, 0, 0, PD_SEQUENTIAL);
		}

        if (ret == POWEROFF || ret == GOTO_MAINMENU){
        	break;
        } else {
        	// ランダム再生続行。ランダム再生順ファイルの次のレコードへ。
        }
	} while (1);
	f_close(&FilRL);
	ActivateSpeaker(0);
	RandomPlaying = 0;

	return ret;
}


/*
 * オプション選択画面
 */
int SelectOption(FIL *Fil){
	int	ret;

	ClearLcd();
    DispMsg(3, MSGNO_SEL_RET, 0, Fil);

	do {
		// １行目で出力先選択
	    if (RecHistory[HS_OUTPUT_OFFSET] == OUTPUT_SPEAKER){
	        DispMsg(0, MSGNO_SPEAKER, 1, Fil);
	    } else {
	        DispMsg(0, MSGNO_EARPHONE, 1, Fil);
	    }

		ret = PauseAndSelectItem();
		if (ret == SWSELECT_PRESSED){
		    if (RecHistory[HS_OUTPUT_OFFSET] == OUTPUT_SPEAKER){
		    	RecHistory[HS_OUTPUT_OFFSET] = OUTPUT_EARPHONE;
		    } else {
		    	RecHistory[HS_OUTPUT_OFFSET] = OUTPUT_SPEAKER;
		    }
		} else if (ret == SWRETURN_PRESSED){
			return TRUE;
		}
	} while (1);
}


/*
 * 履歴ファイルを読み込む
 */
int ReadHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    memset(RecHistory, 0, RECLEN_HISTFILE);
    res = f_open(Fil, MakeUnicodeString(HISTFILE), FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        f_read(Fil, RecHistory, RECLEN_HISTFILE, &br);
        f_close(Fil);
    } else {
    	DEBUG_MSG("ReadHistory f_open error\r\n");
    }

    return TRUE;
}

/*
 * 履歴ファイルを保存
 */
int SaveHistory(FIL *Fil){
    FRESULT     res;
    UINT        br;

    res = f_open(Fil, MakeUnicodeString(HISTFILE), FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
    	//DEBUG_MSG("SaveHistory f_open OK \r\n");
        res = f_write(Fil, RecHistory, RECLEN_HISTFILE, &br);
        if (res == FR_OK){
        	//DEBUG_MSG("SaveHistory f_write OK \r\n");
        }
        f_close(Fil);
        return TRUE;
    } else {
    	DEBUG_MSG("SaveHistory f_open error \r\n");
    }
    return FALSE;
}

/*
 * ディレクトリごとに保存されているVolumeを設定
 */
int SetLocalVolume(const TCHAR *volumefile, int output, FIL *Fil){
    FRESULT     res;
	uint8_t		rec[RECLEN_VOLUME];
    UINT    	br;
    int			i;
    int			err_flg = 0;
    int			idx;
    mad_fixed_t	vol;

    res = f_open(Fil, volumefile, FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
        res = f_read(Fil, rec, RECLEN_VOLUME, &br);
        if (res == FR_OK && br == RECLEN_VOLUME){
        	if (output == OUTPUT_SPEAKER){
        		idx = rec[VF_VOLIDX_SPK_OFFSET];
        		memcpy(&vol, rec + VF_FACTOR_SPK_OFFSET, VF_FACTOR_SPK_LEN);
        	} else {
        		idx = rec[VF_VOLIDX_EAR_OFFSET];
        		memcpy(&vol, rec + VF_FACTOR_EAR_OFFSET, VF_FACTOR_EAR_LEN);
        	}
        	if (vol == mad_f_tofixed((long)VolumeTab[idx]/(float)MAX_VOLUME)){
        		DEBUG_MSG(" idx/vol matched \r\n");
        		VolumeIdx = idx;
        		VolumeFactor = vol;
        	} else {
            	for (i=MIN_VOLUMETAB_IDX; i<(MAX_VOLUMETAB_IDX - 1); i++){
            		if (vol < mad_f_tofixed((long)VolumeTab[i + 1]/(float)MAX_VOLUME)){
            			break;
            		}
            	}
            	VolumeIdx = i;
            	VolumeFactor = mad_f_tofixed((long)VolumeTab[VolumeIdx]/(float)MAX_VOLUME);
        	}
        	DebugWritelnV1(" SetLocalVolume ", VolumeIdx);
        } else {
        	err_flg = 1;
        	//DEBUG_MSG("SetLocalVolume f_read error \r\n");
        }
    	f_close(Fil);
    } else {
    	err_flg = 1;
    	//DEBUG_MSG("SetLocalVolume f_open error \r\n");
    }

    if (err_flg == 1){		// VOLUMEファイルから設定できないときDefault値を設定
    	if (output == OUTPUT_SPEAKER){
    		VolumeIdx = DEF_VOLUMETAB_IDX_SPK;
        	VolumeFactor = mad_f_tofixed((long)VolumeTab[DEF_VOLUMETAB_IDX_SPK]/(float)MAX_VOLUME);
    	} else {
    		VolumeIdx = DEF_VOLUMETAB_IDX_EAR;
        	VolumeFactor = mad_f_tofixed((long)VolumeTab[DEF_VOLUMETAB_IDX_EAR]/(float)MAX_VOLUME);
    	}
    }

    return TRUE;
}


/*
 * ディレクトリごとのVolumeを保存する
 */
int SaveVolumeFile(const TCHAR *volumefile, int output, FIL *Fil){
    FRESULT     res;
	uint8_t		rec[RECLEN_VOLUME];
    UINT    	br;
    mad_fixed_t	vol;

    res = f_open(Fil, volumefile, FA_READ | FA_WRITE | FA_OPEN_ALWAYS);
    if (res == FR_OK){
        res = f_read(Fil, rec, RECLEN_VOLUME, &br);
        if (res != FR_OK || br != RECLEN_VOLUME){		// 保存されていないときはDefault値を初期設定
        	// Earphone
        	rec[VF_VOLIDX_EAR_OFFSET] = DEF_VOLUMETAB_IDX_EAR;
        	vol = mad_f_tofixed((long)VolumeTab[VF_VOLIDX_EAR_OFFSET]/(float)MAX_VOLUME);
    		memcpy(rec + VF_FACTOR_EAR_OFFSET, &vol, VF_FACTOR_EAR_LEN);
    		// Speaker
        	rec[VF_VOLIDX_SPK_OFFSET] = DEF_VOLUMETAB_IDX_SPK;
        	vol = mad_f_tofixed((long)VolumeTab[DEF_VOLUMETAB_IDX_SPK]/(float)MAX_VOLUME);
    		memcpy(rec + VF_FACTOR_SPK_OFFSET, &vol, VF_FACTOR_SPK_LEN);
        }
    	if (output == OUTPUT_SPEAKER){
        	rec[VF_VOLIDX_SPK_OFFSET] = VolumeIdx;
    		memcpy(rec + VF_FACTOR_SPK_OFFSET, (uint8_t *)&VolumeFactor, VF_FACTOR_SPK_LEN);
    	} else {
        	rec[VF_VOLIDX_EAR_OFFSET] = VolumeIdx;
    		memcpy(rec + VF_FACTOR_EAR_OFFSET, (uint8_t *)&VolumeFactor, VF_FACTOR_EAR_LEN);
    	}
		res = f_lseek(Fil, 0);
		res = f_write(Fil, rec, RECLEN_VOLUME, &br);
	    if (res != FR_OK){
	    	DEBUG_MSG("SaveVolumeFile f_write error \r\n");
	    }
    	f_close(Fil);
    	DEBUG_MSG("SaveVolumeFile end \r\n");
    } else {
    	DEBUG_MSG("SaveVolumeFile f_open error \r\n");
    }

    return TRUE;
}


// 選択画面における入力待ち
int PauseAndSelectItem(){
	int		ret;

	//DEBUG_MSG("PauseAndSelectItem");
	ret = 0;
	PinEvent = 0;
	AutoPoffTimer = 0;
	do {
		if (PinEvent != 0){
			ret = PinEvent;
			//DebugWritelnV1(" ret=", ret);
			PinEvent = 0;
			return ret;
		}
		if (AutoPoffTimer == AUTOPOFF_PAUSE){
			return SWPOFF_PRESSED;
		}
	} while (1);
}


/*
 * OLED初期化
 */
void InitOled(){
	int		idx;

	idx = 0;
	I2CMasterBuffer[PORT_USED][idx++] = (OLED_ADDRESS << 1);
	I2CMasterBuffer[PORT_USED][idx++] = OLED_CB_COMMAND;
	I2CMasterBuffer[PORT_USED][idx++] = 0x8D;		// Set charge pump
	I2CMasterBuffer[PORT_USED][idx++] = 0x14;		// Enable charge pump
	I2CMasterBuffer[PORT_USED][idx++] = 0xAF;		// Display ON
	I2CMasterBuffer[PORT_USED][idx++] = 0x20;		// Set memory addressing mode
	I2CMasterBuffer[PORT_USED][idx++] = 0x02;		// Page addressing mode
	I2CMasterBuffer[PORT_USED][idx++] = 0xA1;		// Set Segment Re-map (defaultの上下逆で使うので)
	I2CMasterBuffer[PORT_USED][idx++] = 0xC8;		// Set COM Output Scan Direction : remapped mode (defaultの上下逆で使うので)
	I2CWriteLength[PORT_USED] = idx;
	I2CReadLength[PORT_USED] = 0;
	if (I2CEngine( PORT_USED ) == I2C_TIME_OUT){
	}
}

/*
 * １行分のイメージを表示
 */
void DispLine(int page, uint8_t *imagedata){
	int		i;
	int		idx;

	idx = 0;
	I2CMasterBuffer[PORT_USED][idx++] = (OLED_ADDRESS << 1);
	I2CMasterBuffer[PORT_USED][idx++] = OLED_CB_COMMAND;
	I2CMasterBuffer[PORT_USED][idx++] = 0x20;		// Set memory addressing mode
	I2CMasterBuffer[PORT_USED][idx++] = 0x02;		// Page addressing mode
	I2CMasterBuffer[PORT_USED][idx++] = 0xB0 + page;			// Set Page Start Address
	I2CMasterBuffer[PORT_USED][idx++] = 0x00;
	I2CMasterBuffer[PORT_USED][idx++] = 0x10;
	I2CWriteLength[PORT_USED] = idx;
	I2CReadLength[PORT_USED] = 0;
	if (I2CEngine( PORT_USED ) == I2C_TIME_OUT){
	}

	// すこし待ちが必要
	for (i=0; i<20; i++)
		Pause();

	I2CWriteLength[PORT_USED] = 2 + 128;
	I2CReadLength[PORT_USED] = 0;
	I2CMasterBuffer[PORT_USED][0] = (OLED_ADDRESS << 1);
	I2CMasterBuffer[PORT_USED][1] = OLED_CB_DATA;
	for (i=0; i<128; i++){
		I2CMasterBuffer[PORT_USED][2 + i] = imagedata[i];
	}
	if (I2CEngine( PORT_USED ) == I2C_TIME_OUT){
	}

	// すこし待ちが必要
	for (i=0; i<20; i++)
		Pause();
}

/*
 * LcdImageBuff変数に作成されたイメージデータの反転状態を行単位で変更する。
 * line=0,1,2,3
 */
void SetMenuItemSelected(uint8_t line){
	int x1;

	for (x1=0; x1<OLED_WIDTH; x1++){
		LcdImageBuff[2*line*OLED_WIDTH + x1] = ~LcdImageBuff[2*line*OLED_WIDTH + x1];
		LcdImageBuff[2*line*OLED_WIDTH + OLED_WIDTH + x1] = ~LcdImageBuff[2*line*OLED_WIDTH + OLED_WIDTH + x1];
	}
}

/*
 * LcdImageBuff変数に事前に作成されたイメージをOLED全体にまとめて表示する
 */
void DispLcdFullImage(){
	int		page;
	for (page=0; page<8; page++){
		DispLine(page, LcdImageBuff + page*OLED_WIDTH);
	}
}

/*
 * 全画面クリア
 */
void ClearLcd(){
	memset(LcdImageBuff, 0, LCD_IMAGE_BUFFER_SIZE);
	DispLcdFullImage();
}

/*
 * 固定文字列の表示
 * 	line=0,1,2,3
 * 	msgno=メッセージNo
 * 	inv=1:反転
 */
void DispMsg(int line, int msgno, int inv, FIL *Fil){
    FRESULT     res;
    UINT        br;
	uint8_t		imagedata[MS_IMAGEDATAHALF_LEN];
	int			x1;

    res = f_open(Fil, MakeUnicodeString(MSGSTRFILE), FA_READ | FA_OPEN_EXISTING);
    if (res == FR_OK){
		res = f_lseek(Fil, (DWORD)(msgno) * MS_IMAGEDATA_LEN);
        f_read(Fil, imagedata, MS_IMAGEDATAHALF_LEN, &br);
        if (inv == 1){
        	for (x1=0; x1<MS_IMAGEDATAHALF_LEN; x1++){
        		imagedata[x1] = ~imagedata[x1];
        	}
        }
        DispLine(2*line, imagedata);
        f_read(Fil, imagedata, MS_IMAGEDATAHALF_LEN, &br);
        if (inv == 1){
        	for (x1=0; x1<MS_IMAGEDATAHALF_LEN; x1++){
        		imagedata[x1] = ~imagedata[x1];
        	}
        }
        DispLine(2*line + 1, imagedata);
    }
    f_close(Fil);
}

#if 0
void OledTest(){
    int     i;
    int     page;

//    InitOled();
    page = 0;
	I2CWriteLength[PORT_USED] = 2 + 128;
	I2CReadLength[PORT_USED] = 0;
	I2CMasterBuffer[PORT_USED][0] = (OLED_ADDRESS << 1);
	I2CMasterBuffer[PORT_USED][1] = OLED_CB_DATA;
	for (i=0; i<140; i++){
		I2CMasterBuffer[PORT_USED][2 + i] = i;
	}
	if (I2CEngine( PORT_USED ) == I2C_TIME_OUT){
	}

}
#endif

#if 0
void LFN_play_test(){
    FRESULT     res;
    unsigned char buff[256];
	UINT	bytes;
	static TCHAR test16[] = {0x002F, 0x0054, 0x0045, 0x0053, 0x0054, 0x0031, 0x0036, 0x002E, 0x0044, 0x0041, 0x0054, 0x0000};

//	res = f_open(&FilFL, test16, FA_READ | FA_OPEN_EXISTING);
	res = f_open(&FilFL, "/TEST16.DAT", FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DEBUG_MSG("f_open ERR.\n\r");
		return res;
	}
	DEBUG_MSG("f_open OK.\n\r");
	do {
		res = f_read(&FilFL, buff, 256, &bytes);
		if (res == FR_OK){
		    PlayOneMp3(buff, 0, 0);
		} else break;
	} while (1);
	f_close(&FilFL);
}

void LFN_play_test2(){
    FRESULT     res;
    unsigned char buff[256];
	UINT	bytes;
	static TCHAR test16[] = {0x002F, 0x0054, 0x0045, 0x0053, 0x0054, 0x0031, 0x0036, 0x002E, 0x0044, 0x0041, 0x0054, 0x0000};

	res = f_open(&FilFL2, test16, FA_READ | FA_OPEN_EXISTING);
//	res = f_open(&FilFL, "/TEST16.DAT", FA_READ | FA_OPEN_EXISTING);
	if (res != FR_OK){
		DEBUG_MSG("f_open ERR.\n\r");
		return res;
	}
	DEBUG_MSG("f_open OK.\n\r");
	do {
		res = f_read(&FilFL2, buff, 256, &bytes);
		if (res == FR_OK){
		    PlayOneMp3(buff, 0, 0);
		} else break;
	} while (1);
	f_close(&FilFL2);
}
#endif


/*
 * ASCII文字列のUNICODE文字列化を一時的に作成。
 * 名前をハードコードしているファイルをf_openで開くため。
 */
WCHAR *MakeUnicodeString(char *s){
	static WCHAR	buf[64];
	int		i;
	int		len;

	len = strlen(s);
	for (i=0; i<len; i++){
		buf[i] = s[i];
	}
	buf[len] = 0;
	return buf;
}

void Led2_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 20);
}
void Led3_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 21);
}
void Led4_On(){
	LPC_GPIO1->FIOSET |= (0x01 << 23);
}

void Led2_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 20);
}
void Led3_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 21);
}
void Led4_Off(){
	LPC_GPIO1->FIOCLR |= (0x01 << 23);
}

