/****************************************************************************
 *   $Id:: i2c.c 5865 2010-12-08 21:42:21Z usb00423                         $
 *   Project: NXP LPC17xx I2C example
 *
 *   Description:
 *     This file contains I2C code example which include I2C initialization, 
 *     I2C interrupt handler, and APIs for I2C access.
 *     I2Cサンプルプログラムのソースを移植。I2C1のみをマスター(書き込みだけで読み込みなし)として利用することを前提してソースを整理した。
 *
 ****************************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * NXP Semiconductors assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. NXP Semiconductors
 * reserves the right to make changes in the software without
 * notification. NXP Semiconductors also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
****************************************************************************/
#include "lpc17xx.h"
#include "type.h"
#include "i2c.h"
//#include "mp3_12.h"

#if defined(MP3_16)
volatile uint32_t I2CMasterState[I2C_PORT_NUM] = {I2C_IDLE,I2C_IDLE};
volatile uint32_t timeout[I2C_PORT_NUM] = {0, 0};
#else
volatile uint32_t I2CMasterState[I2C_PORT_NUM] = {I2C_IDLE,I2C_IDLE,I2C_IDLE};
volatile uint32_t timeout[I2C_PORT_NUM] = {0, 0, 0};
#endif

#if defined(MP3_16)
volatile uint32_t I2CCount[I2C_PORT_NUM] = {0, 0};
volatile uint8_t I2CMasterBuffer[I2C_PORT_NUM][I2C_BUFSIZE];
volatile uint8_t I2CSlaveBuffer[I2C_PORT_NUM][I2C_BUFSIZE];
#else
volatile uint8_t I2CMasterBuffer[I2C_PORT_NUM][BUFSIZE];
volatile uint8_t I2CSlaveBuffer[I2C_PORT_NUM][BUFSIZE];
volatile uint32_t I2CCount[I2C_PORT_NUM] = {0, 0, 0};
#endif
volatile uint32_t I2CReadLength[I2C_PORT_NUM];
volatile uint32_t I2CWriteLength[I2C_PORT_NUM];

volatile uint32_t RdIndex0 = 0, RdIndex1 = 0, RdIndex2 = 0;
volatile uint32_t WrIndex0 = 0, WrIndex1 = 0, WrIndex2 = 0;

/* 
From device to device, the I2C communication protocol may vary, 
in the example below, the protocol uses repeated start to read data from or 
write to the device:
For master read: the sequence is: STA,Addr(W),offset,RE-STA,Addr(r),data...STO 
for master write: the sequence is: STA,Addr(W),offset,RE-STA,Addr(w),data...STO
Thus, in state 8, the address is always WRITE. in state 10, the address could 
be READ or WRITE depending on the I2C command.
*/   

/*****************************************************************************
** Function name:		I2C_IRQHandler
**
** Descriptions:		I2C interrupt handler, deal with master mode only.
**
** parameters:			None
** Returned value:		None
** 
*****************************************************************************/
void I2C1_IRQHandler(void) 
{
  uint8_t StatValue;

  timeout[1] = 0;
  /* this handler deals with master read and master write only */
  StatValue = LPC_I2C1->I2STAT;
  switch ( StatValue )
  {
	case 0x08:			/* A Start condition is issued. */
		//onBoardLed1_On();
	WrIndex1 = 0;
	LPC_I2C1->I2DAT = I2CMasterBuffer[1][WrIndex1++];
	LPC_I2C1->I2CONCLR = (I2CONCLR_SIC | I2CONCLR_STAC);
	//onBoardLed1_Off();
	break;
	
	case 0x10:			/* A repeated started is issued */
		//onBoardLed2_On();
	RdIndex1 = 0;
	/* Send SLA with R bit set, */
	LPC_I2C1->I2DAT = I2CMasterBuffer[1][WrIndex1++];
	LPC_I2C1->I2CONCLR = (I2CONCLR_SIC | I2CONCLR_STAC);
	//onBoardLed2_Off();
	break;
	
	case 0x18:			/* Regardless, it's a ACK */
		//onBoardLed3_On();
	if ( I2CWriteLength[1] == 1 )
	{
	  LPC_I2C1->I2CONSET = I2CONSET_STO;      /* Set Stop flag */
	  I2CMasterState[1] = I2C_NO_DATA;
	}
	else
	{
	  LPC_I2C1->I2DAT = I2CMasterBuffer[1][WrIndex1++];
	}
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed3_Off();
	break;

	case 0x28:	/* Data byte has been transmitted, regardless ACK or NACK */
		//onBoardLed4_On();
	if ( WrIndex1 < I2CWriteLength[1] )
	{   
	  LPC_I2C1->I2DAT = I2CMasterBuffer[1][WrIndex1++]; /* this should be the last one */
	}
	else
	{
	  if ( I2CReadLength[1] != 0 )
	  {
		LPC_I2C1->I2CONSET = I2CONSET_STA;	/* Set Repeated-start flag */
	  }
	  else
	  {
		LPC_I2C1->I2CONSET = I2CONSET_STO;      /* Set Stop flag */
		I2CMasterState[1] = I2C_OK;
	  }
	}
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed4_Off();
	break;

	case 0x30:
		//onBoardLed1_On();
	LPC_I2C1->I2CONSET = I2CONSET_STO;      /* Set Stop flag */
	I2CMasterState[1] = I2C_NACK_ON_DATA;
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed1_Off();
	break;
	
	case 0x40:	/* Master Receive, SLA_R has been sent */
		//onBoardLed2_On();
	if ( (RdIndex1 + 1) < I2CReadLength[1] )
	{
	  /* Will go to State 0x50 */
	  LPC_I2C1->I2CONSET = I2CONSET_AA;	/* assert ACK after data is received */
	}
	else
	{
	  /* Will go to State 0x58 */
	  LPC_I2C1->I2CONCLR = I2CONCLR_AAC;	/* assert NACK after data is received */
	}
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed2_Off();
	break;
	
	case 0x50:	/* Data byte has been received, regardless following ACK or NACK */
		//onBoardLed3_On();
	I2CSlaveBuffer[1][RdIndex1++] = LPC_I2C1->I2DAT;		// 読み込みなし。あってもカウントだけ。
	if ( (RdIndex1 + 1) < I2CReadLength[1] )
	{   
	  LPC_I2C1->I2CONSET = I2CONSET_AA;	/* assert ACK after data is received */
	}
	else
	{
	  LPC_I2C1->I2CONCLR = I2CONCLR_AAC;	/* assert NACK on last byte */
	}
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed3_Off();
	break;
	
	case 0x58:
		//onBoardLed4_On();
	I2CSlaveBuffer[1][RdIndex1++] = LPC_I2C1->I2DAT;		// 読み込みなし。あってもカウントだけ。
	I2CMasterState[1] = I2C_OK;
	LPC_I2C1->I2CONSET = I2CONSET_STO;	/* Set Stop flag */
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;	/* Clear SI flag */
	//onBoardLed4_Off();
	break;

	case 0x20:		/* regardless, it's a NACK */
	case 0x48:
		//onBoardLed1_On();
	LPC_I2C1->I2CONSET = I2CONSET_STO;      /* Set Stop flag */
	I2CMasterState[1] = I2C_NACK_ON_ADDRESS;
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed1_Off();
	break;
	
	case 0x38:		/* Arbitration lost, in this example, we don't
					deal with multiple master situation */
	default:
		//onBoardLed2_On();
	I2CMasterState[1] = I2C_ARBITRATION_LOST;
	LPC_I2C1->I2CONCLR = I2CONCLR_SIC;
	//onBoardLed2_Off();
	break;
  }
  return;
}


/*****************************************************************************
** Function name:		I2CStart
**
** Descriptions:		Create I2C start condition, a timeout
**				value is set if the I2C never gets started,
**				and timed out. It's a fatal error. 
**
** parameters:			None
** Returned value:		true or false, return false if timed out
** 
*****************************************************************************/
uint32_t I2CStart( uint32_t portNum )
{
  uint32_t retVal = FALSE;
 
  timeout[portNum] = 0;
  /*--- Issue a start condition ---*/
  LPC_I2C[portNum]->I2CONSET = I2CONSET_STA;	/* Set Start flag */
    
  /*--- Wait until START transmitted ---*/
  while( 1 )
  {
	if ( I2CMasterState[portNum] == I2C_STARTED )
	{
	  retVal = TRUE;
	  break;	
	}
	if ( timeout[portNum] >= MAX_TIMEOUT )
	{
	  retVal = FALSE;
	  break;
	}
	timeout[portNum]++;
  }
  return( retVal );
}

/*****************************************************************************
** Function name:		I2CStop
**
** Descriptions:		Set the I2C stop condition, if the routine
**				never exit, it's a fatal bus error.
**
** parameters:			None
** Returned value:		true or never return
** 
*****************************************************************************/
uint32_t I2CStop( uint32_t portNum )
{
  LPC_I2C[portNum]->I2CONSET = I2CONSET_STO;      /* Set Stop flag */
  LPC_I2C[portNum]->I2CONCLR = I2CONCLR_SIC;  /* Clear SI flag */
            
  /*--- Wait for STOP detected ---*/
  while( LPC_I2C[portNum]->I2CONSET & I2CONSET_STO );
  return TRUE;
}

/*****************************************************************************
** Function name:		I2C1Init
**
** Descriptions:		Initialize I2C controller as a master
**
** parameters:			None
** Returned value:		None
** 
*****************************************************************************/
void I2C1Init( void ) 
{
  LPC_SC->PCONP |= (1 << 19);

#if 1
  /* set PIO0.0 and PIO0.1 to I2C1 SDA and SCL */
  /* function to 11 on both SDA and SCL. */
#if defined(MP3_16)
  LPC_PINCON->PINSEL0 &= ~((0x3<<0)|(0x3<<2));
  LPC_PINCON->PINSEL0 |= ((0x3<<0)|(0x3<<2));		// P0.0->SDA1 P0.1->SCL1
  LPC_PINCON->PINMODE0 &= ~((0x3<<0)|(0x3<<2));
  LPC_PINCON->PINMODE0 |= ((0x2<<0)|(0x2<<2));		/* No pull-up no pull-down */
  LPC_PINCON->PINMODE_OD0 |= ((0x01<<0)|(0x1<<1));	/* Open drain */
  LPC_SC->PCLKSEL1 &= ~(0x3 << 6);
  LPC_SC->PCLKSEL1 |= (0x01 << 6);		// PCLK_I2C1=CCLK
//  LPC_SC->PCLKSEL1 |= (0x00 << 6);		// PCLK_I2C1=CCLK/4
#else		// original
  LPC_PINCON->PINSEL0 &= ~((0x3<<0)|(0x3<<2));
  LPC_PINCON->PINSEL0 |= ((0x3<<0)|(0x3<<2));		// P0.0->SDA1 P0.1->SCL1
  LPC_PINCON->PINMODE0 &= ~((0x3<<0)|(0x3<<2));
  LPC_PINCON->PINMODE0 |= ((0x2<<0)|(0x2<<2));		/* No pull-up no pull-down */
  LPC_PINCON->PINMODE_OD0 |= ((0x01<<0)|(0x1<<1));	/* Open drain */
  LPC_SC->PCLKSEL1 &= (0x3 << 6);
  LPC_SC->PCLKSEL1 |= (0x01 << 6);		// PCLK_I2C1=CCLK
#endif
#endif
#if 0
  /* set PIO0.19 and PIO0.20 to I2C1 SDA and SCL */
  /* function to 11 on both SDA and SCL. */
  LPC_PINCON->PINSEL1 &= ~((0x3<<6)|(0x3<<8));
  LPC_PINCON->PINSEL1 |= ((0x3<<6)|(0x3<<8));
  LPC_PINCON->PINMODE1 &= ~((0x3<<6)|(0x3<<8));
  LPC_PINCON->PINMODE1 |= ((0x2<<6)|(0x2<<8));	/* No pull-up no pull-down */
  LPC_PINCON->PINMODE_OD0 |= ((0x1<<19)|(0x1<<20));	
#endif
 
  /*--- Clear flags ---*/
  LPC_I2C1->I2CONCLR = I2CONCLR_AAC | I2CONCLR_SIC | I2CONCLR_STAC | I2CONCLR_I2ENC;

  /*--- Reset registers ---*/
  LPC_I2C1->I2SCLL   = I2SCLL_SCLL;
  LPC_I2C1->I2SCLH   = I2SCLH_SCLH;

  /* Install interrupt handler */
  NVIC_EnableIRQ(I2C1_IRQn);

  LPC_I2C1->I2CONSET = I2CONSET_I2EN;
  return;
}

/*****************************************************************************
** Function name:		I2CEngine
**
** Descriptions:		The routine to complete a I2C transaction
**						from start to stop. All the intermitten
**						steps are handled in the interrupt handler.
**						Before this routine is called, the read
**						length, write length, I2C master buffer,
**						and I2C command fields need to be filled.
**						see i2cmst.c for more details. 
**
** parameters:			I2C port number
** Returned value:		master state of current I2C port.
** 
*****************************************************************************/
uint32_t I2CEngine( uint32_t portNum ) 
{
  /*--- Issue a start condition ---*/
  LPC_I2C[portNum]->I2CONSET = I2CONSET_STA;	/* Set Start flag */

  I2CMasterState[portNum] = I2C_BUSY;	
//  onBoardLed1_On();

  while ( I2CMasterState[portNum] == I2C_BUSY )
  {
	if ( timeout[portNum] >= MAX_TIMEOUT )
	{
	  I2CMasterState[portNum] = I2C_TIME_OUT;
	  break;
	}
	timeout[portNum]++;
  }
  LPC_I2C[portNum]->I2CONCLR = I2CONCLR_STAC;
//  onBoardLed1_Off();

  return ( I2CMasterState[portNum] );
}

/******************************************************************************
**                            End Of File
******************************************************************************/

